import os
import sys
import traceback
from argparse import ArgumentParser

from album.api import Album
from album.argument_parsing import ArgumentParser as AlbumAP
from album.ci.commandline import (
    configure_repo,
    configure_ssh,
    zenodo_publish,
    zenodo_upload,
    update_index,
    commit_changes,
    merge,
)
from album.ci.controller.release_manager import ReleaseManager
from album.runner import album_logging
from album.runner.album_logging import (
    get_active_logger,
    debug_settings,
    pop_active_logger,
)

module_logger = get_active_logger


def main():
    """Entry points of `album ci`."""
    ci_parser = create_parser()

    module_logger().info("Starting CI release cycle...")
    args = ci_parser.parse_args()

    album_ci_command = ""
    try:
        album_ci_command = sys.argv[
            1
        ]  # album command always expected at second position
    except IndexError:
        ci_parser.error("Please provide a valid action!")

    pop_active_logger()  # there will be a new one with the album instance

    # Makes sure album is initialized.
    album_instance = create_album_instance()
    album_instance.load_or_create_collection()

    album_logging.set_loglevel(args.log)

    release_manager = ReleaseManager(
        album_instance, args.name, args.path, args.src, args.force_retrieve
    )
    module_logger().debug("Running %s command..." % album_ci_command)

    try:
        args.func(release_manager, args)  # execute entry point function
    except Exception as e:
        _handle_exception(e)
    finally:
        release_manager.close()


def _handle_exception(e):
    get_active_logger().error("album-ci command failed: %s" % str(e))
    get_active_logger().debug(traceback.format_exc())
    sys.exit(e)


def create_album_instance() -> Album:
    return Album.Builder().build()


def create_parser():
    parser = AlbumCIParser()

    parser.create_git_command_parser(
        "configure-repo",
        configure_repo,
        "Configures the git configuration for the catalog repository. Sets an name and an email.",
    )

    p = parser.create_catalog_command_parser(
        "configure-ssh",
        configure_ssh,
        "Configures the git push option for the catalog repository to use the ssh protocol instead of https. "
        "Needs the systems git configuration to be configured for ssh usage!",
    )
    p.add_argument(
        "--ci-project-path",
        required=False,
        type=str,
        help="The project path of the git repository."
        'E.g. for https://gitlab.com/album-app/album.git the project path is "album-app/album". '
        "Usually necessary for using album-ci from within a CI-pipeline."
        "Allows switching from https to ssh source."
        'Can be set via environment variable "CI_PROJECT_PATH".',
        default=os.getenv("CI_PROJECT_PATH", None),
    )

    parser.create_zenodo_command_parser(
        "publish",
        zenodo_publish,
        "Publishes the corresponding zenodo deposit of a catalog repository deployment branch.",
    )
    p = parser.create_zenodo_command_parser(
        "upload",
        zenodo_upload,
        "Uploads solution of a catalog repository deployment branch to zenodo."
        "Thereby only allowing a single solution per branch.",
    )
    p.add_argument(
        "--report-file",
        required=False,
        help="Path to a report file. If given, a report of the upload will be created.",
        default="",
        type=str,
    )

    p = parser.create_branch_command_parser(
        "update",
        update_index,
        "Updates the index of the catalog repository to include the solution of a catalog repository deployment branch.",
    )
    p.add_argument(
        "--doi",
        required=False,
        help="Sets the DOI of the solution. Overwrites DOI if already specified in the solution.",
        default="",
        type=str,
    )
    p.add_argument(
        "--deposit-id",
        required=False,
        help="Sets the zenodo deposit_id of the solution.",
        default="",
        type=str,
    )

    parser.create_branch_command_parser(
        "commit",
        commit_changes,
        "Pushes all changes to catalog repository deployment branch to the branch origin.",
    )

    parser.create_pipeline_command_parser(
        "merge", merge, "Merge the updated catalog index from the given branch to main."
    )

    return parser.parser


class AlbumCIParser(AlbumAP):
    def __init__(self):
        super().__init__()
        self.parent_parser = self.create_parent_parser()
        self.parser = self.create_parser()
        self.subparsers = self.parser.add_subparsers(
            title="actions", help="sub-command help"
        )

    @staticmethod
    def create_parent_parser():
        """Parent parser for all subparsers to have the same set of arguments."""
        parent_parser = ArgumentParser(add_help=False)
        # parse logging
        parent_parser.add_argument(
            "--log",
            required=False,
            help="Logging level for your album-ci command. Choose between %s"
            % ", ".join([loglevel.name for loglevel in album_logging.LogLevel]),
            default=album_logging.LogLevel(debug_settings()),
            type=(lambda choice: album_logging.to_loglevel(choice)),
        )
        return parent_parser

    def create_parser(self):
        """Creates the main parser for the album framework."""
        parser = ArgumentParser(
            add_help=True,
            description="album-catalog-admin for managing catalog-deploy requests to a catalog via commandline or CI",
            parents=[self.parent_parser],
        )
        return parser

    def create_catalog_command_parser(
        self, command_name, command_function, command_help
    ):
        """Creates a subparser with all necessary arguments for the catalog deployment management"""
        parser = self.subparsers.add_parser(
            command_name, help=command_help, parents=[self.parent_parser]
        )
        parser.set_defaults(func=command_function)
        parser.add_argument("name", type=str, help="Name of the catalog.")
        parser.add_argument(
            "path",
            type=str,
            help="Path of the catalog repository. Can be an empty folder.",
        )
        parser.add_argument(
            "src",
            type=str,
            help="Source of the catalog. Usually a git repository link.",
        )

        parser.add_argument(
            "--force-retrieve",
            required=False,
            help="If True, download path for the catalog will be force emptied before retrieving the catalog.",
            default=False,
            type=bool,
        )

        return parser

    def create_git_command_parser(self, command_name, command_function, command_help):
        parser = self.create_catalog_command_parser(
            command_name, command_function, command_help
        )
        parser.add_argument(
            "--ci-user-name",
            required=False,
            type=str,
            help="Name to use for all ci operations. "
            'Can be set via environment variable "CI_USER_NAME".'
            "If neither is given, system is required to be proper configured!",
            default=os.getenv("CI_USER_NAME", None),
        )
        parser.add_argument(
            "--ci-user-email",
            required=False,
            type=str,
            help="Email to use for all ci operations. "
            'Can be set via environment variable "CI_USER_EMAIL".'
            "If neither is given, system is required to be proper configured!",
            default=os.getenv("CI_USER_EMAIL", None),
        )

        return parser

    def create_branch_command_parser(
        self, command_name, command_function, command_help
    ):
        parser = self.create_git_command_parser(
            command_name, command_function, command_help
        )
        parser.add_argument(
            "--branch-name",
            required=True,
            type=str,
            help="The branch name of the solution to publish.",
        )
        return parser

    def create_zenodo_command_parser(
        self, command_name, command_function, command_help
    ):
        parser = self.create_branch_command_parser(
            command_name, command_function, command_help
        )
        parser.add_argument(
            "--zenodo-base-url",
            required=False,
            type=str,
            help='The base URL where to upload to. Either "https://sandbox.zenodo.org" or "https://zenodo.org"'
            'Can be set via environment variable "ZENODO_BASE_URL".',
            default=os.getenv("ZENODO_BASE_URL", None),
        )
        parser.add_argument(
            "--zenodo-access-token",
            required=False,
            type=str,
            help="The access token representing the zenodo user account to use. For help see zenodo.org!"
            'Can be set via environment variable "ZENODO_ACCESS_TOKEN".',
            default=os.getenv("ZENODO_ACCESS_TOKEN", None),
        )
        return parser

    def create_pipeline_command_parser(
        self, command_name, command_function, command_help
    ):
        parser = self.create_branch_command_parser(
            command_name, command_function, command_help
        )
        parser.add_argument(
            "--dry-run",
            required=False,
            help="Dry-run option."
            " If this argument is added, commits will not be pushed, only information is shown.",
            action="store_true",
        )
        parser.add_argument(
            "--push-option",
            required=False,
            help="Push options for the catalog repository.",
            default=None,
            nargs="+",
        )
        return parser



import json
import pkgutil
from copy import deepcopy
from pathlib import Path
from tempfile import TemporaryDirectory

from jsonschema import validate, ValidationError
from packaging import version

from album.core.api.controller.controller import IAlbumController
from album.core.api.controller.migration_manager import IMigrationManager
from album.core.api.model.catalog import ICatalog
from album.core.api.model.collection_index import ICollectionIndex
from album.core.model.catalog_index import CatalogIndex
from album.core.model.collection_index import CollectionIndex
from album.runner import album_logging

module_logger = album_logging.get_active_logger


class MigrationManager(IMigrationManager):
    def __init__(self, album: IAlbumController):
        self.schema_solution = None
        self.schema_solution_runner_0_4_2 = None
        self.album = album

    def migrate_collection_index(
        self, collection_index: ICollectionIndex, initial_version
    ):
        self.migrate_catalog_collection_db(
            collection_index.get_path(),
            initial_version,  # current version
            CollectionIndex.version,  # current framework target version
        )

    def _load_catalog_index(self, catalog: ICatalog, initial_version) -> None:
        """Loads current index on disk and migrates if necessary"""
        catalog.load_index()
        self.migrate_catalog_index_db(
            catalog.index().get_path(),
            initial_version,  # current version
            CatalogIndex.version,  # current framework target version
        )

    def migrate_catalog_collection_db(
        self, collection_index_path, curr_version, target_version
    ):
        if curr_version != target_version:
            # todo: execute catalog_collection SQL migration scripts if necessary!
            # todo: set new version in DB
            raise NotImplementedError(
                'Cannot migrate collection from version "%s" to version "%s"!'
                % (curr_version, target_version)
            )
        return collection_index_path

    def migrate_catalog_index_db(
        self, catalog_index_path, curr_version, target_version
    ):
        if curr_version != target_version:
            # todo: execute catalog index SQL migration scripts if necessary!
            # todo: set new version in DB
            raise NotImplementedError(
                "Cannot migrate collection from version %s to version %s."
                % (curr_version, target_version)
            )
        return catalog_index_path

    def load_index(self, catalog: ICatalog):
        with TemporaryDirectory(dir=self.album.configuration().tmp_path()) as tmp_dir:
            catalog.update_index_cache(Path(tmp_dir))
        self._load_catalog_index(catalog, CatalogIndex.version)
        self.album.catalogs().set_version(catalog)

    def refresh_index(self, catalog: ICatalog) -> bool:
        with TemporaryDirectory(dir=self.album.configuration().tmp_path()) as tmp_dir:
            if catalog.update_index_cache_if_possible(tmp_dir):
                self._load_catalog_index(catalog, CatalogIndex.version)
                return True
        return False

    def migrate_solution_attrs(self, attrs: dict) -> dict:
        self._load_solution_schema()
        if "album_api_version" not in attrs:
            raise ValidationError(
                "Your setup method is missing the 'album_api_version' keyword - the most recent "
                "version known to this album installation is '0.5.1'."
            )
        api_version = version.parse(attrs["album_api_version"])
        if api_version >= version.parse("0.5.1"):
            validate(attrs, self.schema_solution)
            return attrs
        else:
            validate(attrs, self.schema_solution_runner_0_4_2)
            return self._convert_schema0_schema1(attrs)

    def _load_solution_schema(self):
        if not self.schema_solution:
            data = pkgutil.get_data("album.runner.core.schema", "solution_schema.json")
            self.schema_solution = json.loads(data)
        if not self.schema_solution_runner_0_4_2:
            data = pkgutil.get_data("album.core.schema", "solution_schema_0.4.2.json")
            self.schema_solution_runner_0_4_2 = json.loads(data)

    @staticmethod
    def _convert_schema0_schema1(attrs):
        if "authors" in attrs:
            attrs["solution_creators"] = deepcopy(attrs["authors"])
            attrs.pop("authors")
        return attrs



import pkgutil
from datetime import datetime
from typing import Optional, List

from album.core.api.model.collection_index import ICollectionIndex
from album.core.model.catalog_index import CatalogIndex
from album.core.model.database import Database
from album.core.utils.operations.file_operations import get_dict_entry
from album.core.utils.operations.solution_operations import get_solution_hash
from album.runner.core.api.model.coordinates import ICoordinates


class CollectionIndex(ICollectionIndex, Database):
    version = "0.1.0"

    class CollectionSolution(ICollectionIndex.ICollectionSolution):
        def __init__(self, setup: dict = None, internal: dict = None):
            if setup:
                self._setup = setup
            else:
                self._setup = {}
            if internal:
                self._internal = internal
            else:
                self._internal = {}

        def __eq__(self, other) -> bool:
            return other.setup() == self._setup and other.internal() == self._internal

        def setup(self) -> dict:
            return self._setup

        def internal(self) -> dict:
            return self._internal

    def __init__(self, name, path):
        self.name = name
        super().__init__(path)

    def create(self):
        data = pkgutil.get_data("album.core.schema", "catalog_collection_schema.sql")
        cursor = self.get_cursor()
        cursor.executescript(data.decode("utf-8"))
        self.update_name_version(self.name, self.version, close=False)

        self.close_current_connection()

    def update_name_version(self, name, version, close=True):
        curr_name = self.get_name(close=False)

        cursor = self.get_cursor()
        if curr_name:
            cursor.execute(
                "UPDATE catalog_collection SET name=:name, version=:version WHERE name_id=:name_id",
                {"name_id": 1, "name": name, "version": version},
            )
        else:
            cursor.execute(
                "INSERT INTO catalog_collection values (?, ?, ?)", (1, name, version)
            )

        if close:
            self.close_current_connection()

    def get_name(self, close=True):
        cursor = self.get_cursor()

        r = cursor.execute("SELECT * FROM catalog_collection").fetchone()

        cur_name = r["name"] if r else None

        if close:
            self.close_current_connection()

        return cur_name

    def get_version(self, close=True):
        cursor = self.get_cursor()

        r = cursor.execute("SELECT * FROM catalog_collection").fetchone()

        cur_version = r["version"] if r else None

        if close:
            self.close_current_connection()

        return cur_version

    def is_empty(self, close=True):
        cursor = self.get_cursor()

        r = cursor.execute("SELECT * FROM collection").fetchone()

        if close:
            self.close_current_connection()

        return False if r else True

    # ### catalog ###

    def insert_catalog(
        self, name, src, path, deletable, branch_name, catalog_type, close=True
    ):
        next_id = self.next_id("catalog")
        cursor = self.get_cursor()

        cursor.execute(
            "INSERT INTO catalog VALUES (?, ?, ?, ?, ?, ?, ?)",
            (next_id, name, src, path, branch_name, catalog_type, deletable),
        )

        if close:
            self.close_current_connection()

        return next_id

    def get_catalog(self, catalog_id, close=True):
        cursor = self.get_cursor()

        r = cursor.execute(
            "SELECT * FROM catalog WHERE catalog_id=:catalog_id",
            {
                "catalog_id": catalog_id,
            },
        ).fetchone()

        catalog = None
        if r:
            catalog = dict(r)

        if close:
            self.close_current_connection()

        return catalog

    def get_catalog_by_name(self, catalog_name: str, close=True):
        cursor = self.get_cursor()

        r = cursor.execute(
            "SELECT * FROM catalog WHERE name=:catalog_name",
            {
                "catalog_name": catalog_name,
            },
        ).fetchone()

        catalog = None
        if r:
            catalog = dict(r)

        if close:
            self.close_current_connection()

        return catalog

    def get_catalog_by_path(self, catalog_path: str, close=True):
        cursor = self.get_cursor()

        r = cursor.execute(
            "SELECT * FROM catalog WHERE path=:catalog_path",
            {
                "catalog_path": catalog_path,
            },
        ).fetchone()

        catalog = None
        if r:
            catalog = dict(r)

        if close:
            self.close_current_connection()

        return catalog

    def get_catalog_by_src(self, catalog_src: str, close=True):
        cursor = self.get_cursor()

        r = cursor.execute(
            "SELECT * FROM catalog WHERE src=:catalog_src",
            {
                "catalog_src": catalog_src,
            },
        ).fetchone()

        catalog = None
        if r:
            catalog = dict(r)

        if close:
            self.close_current_connection()

        return catalog

    def get_all_catalogs(self, close=True):
        catalog_list = []
        cursor = self.get_cursor()
        for row in cursor.execute("SELECT * FROM catalog"):
            catalog_list.append(dict(row))

        if close:
            self.close_current_connection()

        return catalog_list

    def remove_catalog(self, catalog_id, close=True):
        cursor = self.get_cursor()
        cursor.execute(
            "DELETE FROM collection " "WHERE catalog_id=:catalog_id",
            {"catalog_id": catalog_id},
        )
        cursor.execute(
            "DELETE FROM collection_tag " "WHERE catalog_id=:catalog_id",
            {"catalog_id": catalog_id},
        )
        cursor.execute(
            "DELETE FROM tag " "WHERE catalog_id=:catalog_id",
            {"catalog_id": catalog_id},
        )
        cursor.execute(
            "DELETE FROM collection_author " "WHERE catalog_id=:catalog_id",
            {"catalog_id": catalog_id},
        )
        cursor.execute(
            "DELETE FROM author " "WHERE catalog_id=:catalog_id",
            {"catalog_id": catalog_id},
        )
        cursor.execute(
            "DELETE FROM collection_citation " "WHERE catalog_id=:catalog_id",
            {"catalog_id": catalog_id},
        )
        cursor.execute(
            "DELETE FROM citation " "WHERE catalog_id=:catalog_id",
            {"catalog_id": catalog_id},
        )
        cursor.execute(
            "DELETE FROM collection_argument " "WHERE catalog_id=:catalog_id",
            {"catalog_id": catalog_id},
        )
        cursor.execute(
            "DELETE FROM argument " "WHERE catalog_id=:catalog_id",
            {"catalog_id": catalog_id},
        )
        cursor.execute(
            "DELETE FROM collection_custom " "WHERE catalog_id=:catalog_id",
            {"catalog_id": catalog_id},
        )
        cursor.execute(
            "DELETE FROM custom " "WHERE catalog_id=:catalog_id",
            {"catalog_id": catalog_id},
        )
        cursor.execute(
            "DELETE FROM cover " "WHERE catalog_id=:catalog_id",
            {"catalog_id": catalog_id},
        )
        cursor.execute(
            "DELETE FROM documentation " "WHERE catalog_id=:catalog_id",
            {"catalog_id": catalog_id},
        )
        cursor.execute(
            "DELETE FROM collection_collection " "WHERE catalog_id_parent=:catalog_id",
            {"catalog_id": catalog_id},
        )
        cursor.execute(
            "DELETE FROM collection_collection " "WHERE catalog_id_child=:catalog_id",
            {"catalog_id": catalog_id},
        )
        cursor.execute(
            "DELETE FROM catalog " "WHERE catalog_id=:catalog_id",
            {"catalog_id": catalog_id},
        )

        if close:
            self.close_current_connection()

    # ### collection ###

    def insert_solution(self, catalog_id, solution_attrs, close=True):
        collection_id = self.next_id("collection")
        hash_val = get_dict_entry(solution_attrs, "hash", allow_none=True)

        # there must be a hash value
        if not hash_val:
            hash_val = get_solution_hash(
                solution_attrs, CatalogIndex.get_solution_column_keys()
            )

        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO collection VALUES "
            "(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ? ,? ,?, ?, ? ,?, ?, ?, ?)",
            (
                collection_id,
                get_dict_entry(solution_attrs, "solution_id"),
                solution_attrs["group"],
                solution_attrs["name"],
                get_dict_entry(solution_attrs, "title"),
                solution_attrs["version"],
                get_dict_entry(solution_attrs, "timestamp"),
                get_dict_entry(solution_attrs, "description"),
                get_dict_entry(solution_attrs, "doi"),
                get_dict_entry(solution_attrs, "license"),
                get_dict_entry(solution_attrs, "album_version"),
                get_dict_entry(solution_attrs, "album_api_version"),
                get_dict_entry(solution_attrs, "changelog"),
                get_dict_entry(solution_attrs, "acknowledgement"),
                hash_val,
                None,  # when installed?
                None,  # last executed
                0,  # installation unfinished
                0,  # installed
                catalog_id,
            ),
        )

        if "solution_creators" in solution_attrs:
            for author in solution_attrs["solution_creators"]:
                author_id = self._exists_author(author, catalog_id, close=False)
                if not author_id:
                    author_id = self._insert_author(author, catalog_id, close=False)
                self._insert_collection_author(
                    collection_id, author_id, catalog_id, close=False
                )

        if "tags" in solution_attrs:
            for tag in solution_attrs["tags"]:
                tag_id = self._exists_tag(tag, catalog_id, close=False)
                if not tag_id:
                    tag_id = self._insert_tag(tag, catalog_id, close=False)
                self._insert_collection_tag(
                    collection_id, tag_id, catalog_id, close=False
                )

        if "cite" in solution_attrs:
            for citation in solution_attrs["cite"]:
                citation_id = self._exists_citation(citation, catalog_id, close=False)
                if not citation_id:
                    citation_id = self._insert_citation(
                        citation, catalog_id, close=False
                    )
                self._insert_collection_citation(
                    collection_id, citation_id, catalog_id, close=False
                )

        if "args" in solution_attrs:
            for argument in solution_attrs["args"]:
                argument_id = self._exists_argument(argument, catalog_id)
                if not argument_id:
                    argument_id = self._insert_argument(
                        argument, catalog_id, close=False
                    )
                self._insert_collection_argument(
                    collection_id, argument_id, catalog_id, close=False
                )

        if "covers" in solution_attrs:
            for cover in solution_attrs["covers"]:
                if not self._exists_cover(cover, catalog_id, collection_id):
                    self._insert_cover(cover, catalog_id, collection_id, close=False)

        if "documentation" in solution_attrs:
            for documentation in solution_attrs["documentation"]:
                if not self._exists_documentation(
                    documentation, catalog_id, collection_id, close=False
                ):
                    self._insert_documentation(
                        documentation, catalog_id, collection_id, close=False
                    )

        if "custom" in solution_attrs:
            for custom in solution_attrs["custom"]:
                custom_value = solution_attrs["custom"][custom]
                custom_id = self._exists_custom(custom, custom_value, catalog_id)
                if not custom_id:
                    custom_id = self._insert_custom(
                        custom, custom_value, catalog_id, close=False
                    )
                self._insert_collection_custom(
                    collection_id, custom_id, catalog_id, close=False
                )

        if close:
            self.close_current_connection()

        return collection_id

    def _insert_collection_argument(
        self, collection_id, argument_id, catalog_id, close=True
    ):
        collection_solution_argument_id = self.next_id("collection_argument")

        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO collection_argument values (?, ?, ?, ?)",
            (collection_solution_argument_id, collection_id, argument_id, catalog_id),
        )

        if close:
            self.close_current_connection()

        return collection_solution_argument_id

    def _insert_collection_custom(
        self, collection_id, custom_id, catalog_id, close=True
    ):
        collection_solution_custom_id = self.next_id("collection_custom")

        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO collection_custom values (?, ?, ?, ?)",
            (collection_solution_custom_id, collection_id, custom_id, catalog_id),
        )

        if close:
            self.close_current_connection()

        return collection_solution_custom_id

    def _insert_collection_citation(
        self, collection_id, citation_id, catalog_id, close=True
    ):
        collection_solution_citation_id = self.next_id("collection_citation")

        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO collection_citation values (?, ?, ?, ?)",
            (collection_solution_citation_id, collection_id, citation_id, catalog_id),
        )

        if close:
            self.close_current_connection()

        return collection_solution_citation_id

    def _insert_collection_tag(self, collection_id, tag_id, catalog_id, close=True):
        collection_solution_tag_id = self.next_id("collection_tag")

        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO collection_tag values (?, ?, ?, ?)",
            (collection_solution_tag_id, collection_id, tag_id, catalog_id),
        )

        if close:
            self.close_current_connection()

        return collection_solution_tag_id

    def _insert_collection_author(
        self, collection_id, author_id, catalog_id, close=True
    ):
        collection_author_id = self.next_id("collection_author")

        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO collection_author values (?, ?, ?, ?)",
            (collection_author_id, collection_id, author_id, catalog_id),
        )

        if close:
            self.close_current_connection()

        return collection_author_id

    def _exists_author(self, author_name, catalog_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM author WHERE name=:author_name AND catalog_id=:catalog_id",
            {"author_name": author_name, "catalog_id": catalog_id},
        ).fetchone()

        if close:
            self.close_current_connection()

        return r["author_id"] if r else None

    def _insert_author(self, author, catalog_id, close=True):
        author_id = self.next_id("author")

        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO author values (?, ?, ?)", (author_id, catalog_id, author)
        )

        if close:
            self.close_current_connection()

        return author_id

    def _exists_tag(self, tag_name, catalog_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM tag WHERE name=:tag_name AND catalog_id=:catalog_id",
            {"tag_name": tag_name, "catalog_id": catalog_id},
        ).fetchone()

        if close:
            self.close_current_connection()

        return r["tag_id"] if r else None

    def _insert_tag(self, tag, catalog_id, close=True):
        tag_id = self.next_id("tag")
        cursor = self.get_cursor()

        cursor.execute(
            "INSERT INTO tag values (?, ?, ?, ?)", (tag_id, catalog_id, tag, "manual")
        )

        if close:
            self.close_current_connection()

        return tag_id

    def _exists_argument(self, argument, catalog_id, close=True):
        cursor = self.get_cursor()

        exc_str = (
            "SELECT * FROM argument WHERE name=:argument_name AND description=:argument_description "
            "AND catalog_id=:catalog_id "
        )
        exc_val = {
            "argument_name": argument["name"],
            "argument_description": argument["description"],
            "catalog_id": catalog_id,
        }
        argument_type = get_dict_entry(argument, "type")
        argument_default_value = get_dict_entry(argument, "default")

        if argument_type:
            exc_str += "AND type=:argument_type "
            exc_val["argument_type"] = argument_type
        else:
            exc_str += "AND type IS NULL "

        if argument_default_value:
            exc_str += "AND default_value=:argument_default_value "
            exc_val["argument_default_value"] = argument_default_value
        else:
            exc_str += "AND default_value IS NULL "

        r = cursor.execute(exc_str, exc_val).fetchone()

        if close:
            self.close_current_connection()

        return r["argument_id"] if r else None

    def _exists_custom(self, custom_key, custom_value, catalog_id, close=True):
        cursor = self.get_cursor()

        exc_str = (
            "SELECT * FROM custom WHERE custom_key=:custom_key AND custom_value=:custom_value "
            "AND catalog_id=:catalog_id "
        )
        exc_val = {
            "custom_key": custom_key,
            "custom_value": custom_value,
            "catalog_id": catalog_id,
        }

        r = cursor.execute(exc_str, exc_val).fetchone()

        if close:
            self.close_current_connection()

        return r["custom_id"] if r else None

    def _insert_argument(self, argument, catalog_id, close=True):
        argument_id = self.next_id("argument")
        cursor = self.get_cursor()

        cursor.execute(
            "INSERT INTO argument values (?, ?, ?, ?, ?, ?, ?)",
            (
                argument_id,
                catalog_id,
                argument["name"],
                get_dict_entry(argument, "type"),
                argument["description"],
                get_dict_entry(argument, "default"),
                get_dict_entry(argument, "required"),
            ),
        )

        if close:
            self.close_current_connection()

        return argument_id

    def _insert_custom(self, custom_key, custom_value, catalog_id, close=True):
        custom_id = self.next_id("custom")
        cursor = self.get_cursor()

        cursor.execute(
            "INSERT INTO custom values (?, ?, ?, ?)",
            (custom_id, catalog_id, custom_key, custom_value),
        )

        if close:
            self.close_current_connection()

        return custom_id

    def _exists_citation(self, citation, catalog_id, close=True):
        cursor = self.get_cursor()

        exc_str = "SELECT * FROM citation WHERE text=:citation_text AND catalog_id=:catalog_id "
        exc_val = {"citation_text": citation["text"], "catalog_id": catalog_id}
        citation_doi = get_dict_entry(citation, "doi")

        if citation_doi:
            exc_str += "AND doi=:citation_doi"
            exc_val["citation_doi"] = citation_doi
        else:
            exc_str += "AND doi IS NULL"

        r = cursor.execute(exc_str, exc_val).fetchone()

        if close:
            self.close_current_connection()

        return r["citation_id"] if r else None

    def _insert_citation(self, citation, catalog_id, close=True):
        citation_id = self.next_id("citation")

        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO citation values (?, ?, ?, ?, ?)",
            (
                citation_id,
                catalog_id,
                citation["text"],
                get_dict_entry(citation, "doi"),
                get_dict_entry(citation, "url"),
            ),
        )

        if close:
            self.close_current_connection()

        return citation_id

    def _exists_cover(self, cover, catalog_id, collection_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM cover WHERE source=:cover_source AND description=:cover_description "
            "AND catalog_id=:catalog_id AND collection_id=:collection_id",
            {
                "cover_source": cover["source"],
                "cover_description": cover["description"],
                "catalog_id": catalog_id,
                "collection_id": collection_id,
            },
        ).fetchone()

        if close:
            self.close_current_connection()

        return r["cover_id"] if r else None

    def _insert_cover(self, cover, catalog_id, collection_id, close=True):
        cover_id = self.next_id("cover")

        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO cover values (?, ?, ?, ?, ?)",
            (
                cover_id,
                collection_id,
                catalog_id,
                cover["source"],
                cover["description"],
            ),
        )

        if close:
            self.close_current_connection()

        return cover_id

    def _exists_documentation(
        self, documentation, catalog_id, collection_id, close=True
    ):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM main.documentation WHERE documentation=:documentation "
            "AND catalog_id=:catalog_id AND collection_id=:collection_id",
            {
                "documentation": documentation,
                "catalog_id": catalog_id,
                "collection_id": collection_id,
            },
        ).fetchone()

        if close:
            self.close_current_connection()

        return r["documentation_id"] if r else None

    def _insert_documentation(
        self, documentation, catalog_id, collection_id, close=True
    ):
        documentation_id = self.next_id("documentation")
        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO documentation values (?, ?, ?, ?)",
            (documentation_id, collection_id, catalog_id, documentation),
        )

        if close:
            self.close_current_connection()

        return documentation_id

    def insert_collection_collection(
        self,
        collection_id_parent,
        collection_id_child,
        catalog_id_parent,
        catalog_id_child,
        close=True,
    ):

        if not self._exists_collection_collection(
            collection_id_parent,
            collection_id_child,
            catalog_id_parent,
            catalog_id_child,
            close=False,
        ):

            solution_solution_id = self.next_id("collection_collection")

            cursor = self.get_cursor()
            cursor.execute(
                "INSERT INTO collection_collection values (?, ?, ?, ?, ?)",
                (
                    solution_solution_id,
                    collection_id_parent,
                    collection_id_child,
                    catalog_id_parent,
                    catalog_id_child,
                ),
            )
        if close:
            self.close_current_connection()

    def _exists_collection_collection(
        self,
        collection_id_parent,
        collection_id_child,
        catalog_id_parent,
        catalog_id_child,
        close=True,
    ):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM collection_collection WHERE collection_id_parent=:collection_id_parent "
            "AND collection_id_child=:collection_id_child AND catalog_id_parent=:catalog_id_parent "
            "AND catalog_id_child=:catalog_id_child",
            {
                "collection_id_parent": collection_id_parent,
                "collection_id_child": collection_id_child,
                "catalog_id_parent": catalog_id_parent,
                "catalog_id_child": catalog_id_child,
            },
        ).fetchone()

        if close:
            self.close_current_connection()

        return r["collection_collection_id"] if r else None

    def remove_parent(self, collection_id, close=True):
        cursor = self.get_cursor()
        cursor.execute(
            "DELETE FROM collection_collection WHERE collection_id_child=:collection_id",
            {"collection_id": collection_id},
        )

        if close:
            self.close_current_connection()

    def _get_children_of_solution(self, collection_id, close=True):
        child_solutions = []
        cursor = self.get_cursor()

        r = cursor.execute(
            "SELECT css.* FROM collection_collection css "
            "JOIN collection c ON c.collection_id = css.collection_id_parent "
            "WHERE c.collection_id=:collection_id",
            {"collection_id": collection_id},
        ).fetchall()

        for row in r:
            child_solutions.append(
                dict(row)
            )  # do not resolve get this solution here: recursion!

        if close:
            self.close_current_connection()

        return child_solutions

    def get_parent_of_solution(
        self, collection_id, close=True
    ) -> Optional[CollectionSolution]:
        parent_solution = None
        cursor = self.get_cursor()

        r = cursor.execute(
            "SELECT css.* FROM collection_collection css "
            "JOIN collection c ON c.collection_id = css.collection_id_child "
            "WHERE c.collection_id=:collection_id",
            {"collection_id": collection_id},
        ).fetchall()

        if len(r) > 1:
            raise KeyError(
                "Database error. Solution with id %s has several parents!"
                % str(collection_id)
            )

        for row in r:
            parent_solution = self.get_solution_by_collection_id(
                row["collection_id_parent"], close=False
            )

        if close:
            self.close_current_connection()

        return parent_solution

    def get_all_solutions(self, close=True) -> List[CollectionSolution]:
        solutions_list = []
        cursor = self.get_cursor()
        for row in cursor.execute("SELECT * FROM collection").fetchall():
            solution = self._process_solution_row(dict(row), close=False)
            solutions_list.append(solution)

        if close:
            self.close_current_connection()

        return solutions_list

    def get_all_installed_solutions_by_catalog(self, catalog_id, close=True):
        solutions_list = []

        cursor = self.get_cursor()
        for row in cursor.execute(
            "SELECT * FROM collection WHERE installed=:installed AND catalog_id=:catalog_id",
            {"installed": 1, "catalog_id": catalog_id},
        ).fetchall():
            solution = self._process_solution_row(dict(row), close=False)
            solutions_list.append(solution)

        if close:
            self.close_current_connection()

        return solutions_list

    def _process_solution_row(self, solution_dict, close=True) -> CollectionSolution:
        setup = {}
        internal = {}
        collection_id = solution_dict["collection_id"]
        for key in CatalogIndex.get_solution_column_keys():
            setup[key] = solution_dict[key]
        for key in solution_dict.keys():
            if key not in setup:
                internal[key] = solution_dict[key]
        setup["solution_creators"] = self._get_authors_by_solution(
            collection_id, close=False
        )
        setup["tags"] = self._get_tags_by_solution(collection_id, close=False)
        setup["cite"] = self._get_citations_by_solution(collection_id, close=False)
        setup["args"] = self._get_arguments_by_solution(collection_id, close=False)
        setup["covers"] = self._get_covers_by_solution(collection_id, close=False)
        setup["documentation"] = self._get_documentation_by_solution(
            collection_id, close=False
        )
        setup["custom"] = self._get_custom_by_solution(collection_id, close=False)
        internal["children"] = self._get_children_of_solution(
            collection_id, close=False
        )
        internal["parent"] = self.get_parent_of_solution(collection_id, close=close)
        return CollectionIndex.CollectionSolution(setup, internal)

    def _get_authors_by_solution(self, collection_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT a.* FROM author a "
            "JOIN collection_author sa ON sa.author_id = a.author_id "
            "WHERE sa.collection_id=:collection_id",
            {"collection_id": collection_id},
        ).fetchall()

        res = []
        for row in r:
            res.append(row["name"])

        if close:
            self.close_current_connection()

        return res

    def _get_arguments_by_solution(self, collection_id, close=True):
        cursor = self.get_cursor()

        r = cursor.execute(
            "SELECT a.* FROM argument a "
            "JOIN collection_argument sa ON sa.argument_id = a.argument_id "
            "WHERE sa.collection_id=:collection_id",
            {"collection_id": collection_id},
        ).fetchall()

        res = []
        for row in r:
            row = dict(row)
            argument = {"name": row["name"], "type": row["type"]}
            if "description" in row and row["description"] is not None:
                argument["description"] = row["description"]
            if "required" in row and row["required"] is not None:
                argument["required"] = bool(row["required"])
            if "default_value" in row and row["default_value"] is not None:
                argument["default"] = row["default_value"]
            res.append(argument)

        if close:
            self.close_current_connection()

        return res

    def _get_custom_by_solution(self, collection_id, close=True):
        cursor = self.get_cursor()

        r = cursor.execute(
            "SELECT a.* FROM custom a "
            "JOIN collection_custom sa ON sa.custom_id = a.custom_id "
            "WHERE sa.collection_id=:collection_id",
            {"collection_id": collection_id},
        ).fetchall()

        res = {}
        for row in r:
            res[row["custom_key"]] = row["custom_value"]

        if close:
            self.close_current_connection()

        return res

    def _get_tags_by_solution(self, collection_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT t.* FROM tag t "
            "JOIN collection_tag st ON st.tag_id = t.tag_id "
            "WHERE st.collection_id=:collection_id",
            {"collection_id": collection_id},
        ).fetchall()

        res = []
        for row in r:
            res.append(row["name"])

        if close:
            self.close_current_connection()

        return res

    def _get_citations_by_solution(self, collection_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT c.* FROM citation c "
            "JOIN collection_citation sc ON sc.citation_id = c.citation_id "
            "WHERE sc.collection_id=:collection_id",
            {"collection_id": collection_id},
        ).fetchall()

        res = []
        for row in r:
            citation = {"text": row["text"]}
            if row["doi"]:
                citation["doi"] = row["doi"]
            if row["url"]:
                citation["url"] = row["url"]
            res.append(citation)

        if close:
            self.close_current_connection()

        return res

    def _get_covers_by_solution(self, collection_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT c.* FROM cover c " "WHERE c.collection_id=:collection_id",
            {"collection_id": collection_id},
        ).fetchall()

        res = []
        for row in r:
            cover = {"description": row["description"], "source": row["source"]}
            res.append(cover)

        if close:
            self.close_current_connection()

        return res

    def _get_documentation_by_solution(self, collection_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT d.* FROM documentation d " "WHERE d.collection_id=:collection_id",
            {"collection_id": collection_id},
        ).fetchall()

        res = []
        for row in r:
            res.append(row["documentation"])

        if close:
            self.close_current_connection()

        return res

    def get_solutions_by_catalog(
        self, catalog_id, close=True
    ) -> List[CollectionSolution]:
        catalog_solutions = []

        cursor = self.get_cursor()
        for row in cursor.execute(
            "SELECT c.* FROM collection c " "WHERE c.catalog_id=:catalog_id",
            {"catalog_id": catalog_id},
        ).fetchall():
            solution = self._process_solution_row(dict(row), close=False)
            catalog_solutions.append(solution)

        if close:
            self.close_current_connection()

        return catalog_solutions

    def get_solution_by_hash(
        self, hash_value, close=True
    ) -> Optional[CollectionSolution]:
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM collection WHERE hash=:hash_value",
            {"hash_value": hash_value},
        ).fetchone()

        solution = None
        if r:
            solution = self._process_solution_row(dict(r), close=False)

        if close:
            self.close_current_connection()

        return solution

    def get_solution_by_collection_id(
        self, collection_id, close=True
    ) -> Optional[CollectionSolution]:
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM collection WHERE collection_id=:collection_id",
            {
                "collection_id": collection_id,
            },
        ).fetchone()

        solution = None
        if r:
            solution = self._process_solution_row(dict(r), close=False)

        if close:
            self.close_current_connection()

        return solution

    def get_solution_by_doi(self, doi, close=True) -> Optional[CollectionSolution]:
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM collection WHERE doi=:doi",
            {
                "doi": doi,
            },
        ).fetchone()

        solution = None
        if r:
            solution = self._process_solution_row(dict(r), close=False)

        if close:
            self.close_current_connection()

        return solution

    def get_solution_by_catalog_grp_name_version(
        self, catalog_id, coordinates: ICoordinates, close=True
    ) -> Optional[CollectionSolution]:
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM collection "
            'WHERE catalog_id=:catalog_id AND "group"=:group AND name=:name AND version=:version',
            {
                "catalog_id": catalog_id,
                "group": coordinates.group(),
                "name": coordinates.name(),
                "version": coordinates.version(),
            },
        ).fetchall()

        if len(r) > 1:
            raise KeyError(
                "Database error. Please reinstall the solution %s from catalog %s !"
                % (coordinates.group(), catalog_id)
            )

        solution = None
        for row in r:
            solution = self._process_solution_row(dict(row), close=False)

        if close:
            self.close_current_connection()

        return solution

    def get_solutions_by_grp_name_version(
        self, coordinates: ICoordinates, close=True
    ) -> List[CollectionSolution]:
        solutions_list = []

        cursor = self.get_cursor()
        for row in cursor.execute(
            'SELECT * FROM collection WHERE "group"=:group AND name=:name AND version=:version',
            {
                "group": coordinates.group(),
                "name": coordinates.name(),
                "version": coordinates.version(),
            },
        ).fetchall():
            solution = self._process_solution_row(dict(row), close=False)
            solutions_list.append(solution)

        if close:
            self.close_current_connection()

        return solutions_list

    def get_solutions_by_grp_name(
        self, group: str, name: str, close=True
    ) -> List[CollectionSolution]:
        solutions_list = []

        cursor = self.get_cursor()
        for row in cursor.execute(
            'SELECT * FROM collection WHERE "group"=:group AND name=:name',
            {"group": group, "name": name},
        ).fetchall():
            solution = self._process_solution_row(dict(row), close=False)
            solutions_list.append(solution)

        if close:
            self.close_current_connection()

        return solutions_list

    def get_solutions_by_name_version(
        self, name: str, version: str, close=True
    ) -> List[CollectionSolution]:
        solutions_list = []

        cursor = self.get_cursor()
        for row in cursor.execute(
            "SELECT * FROM collection WHERE name=:name AND version=:version",
            {
                "name": name,
                "version": version,
            },
        ).fetchall():
            solution = self._process_solution_row(dict(row), close=False)
            solutions_list.append(solution)

        if close:
            self.close_current_connection()

        return solutions_list

    def get_solutions_by_name(self, name: str, close=True) -> List[CollectionSolution]:
        solutions_list = []

        cursor = self.get_cursor()
        for row in cursor.execute(
            "SELECT * FROM collection WHERE name=:name", {"name": name}
        ).fetchall():
            solution = self._process_solution_row(dict(row), close=False)
            solutions_list.append(solution)

        if close:
            self.close_current_connection()

        return solutions_list

    def get_recently_installed_solutions(self, close=True) -> List[CollectionSolution]:
        solutions_list = []

        cursor = self.get_cursor()
        for row in cursor.execute(
            "SELECT * FROM collection WHERE installed=:installed ORDER BY install_date ",
            {"installed": 1},
        ).fetchall():
            solution = self._process_solution_row(dict(row), close=False)
            solutions_list.append(solution)

        if close:
            self.close_current_connection()

        return solutions_list

    def get_recently_launched_solutions(self, close=True) -> List[CollectionSolution]:
        solutions_list = []

        cursor = self.get_cursor()
        for row in cursor.execute(
            "SELECT * FROM collection WHERE last_execution IS NOT NULL ORDER BY last_execution"
        ).fetchall():
            solution = self._process_solution_row(dict(row), close=False)
            solutions_list.append(solution)

        if close:
            self.close_current_connection()

        return solutions_list

    def get_unfinished_installation_solutions(self, close=True):
        solutions_list = []

        cursor = self.get_cursor()
        for row in cursor.execute(
            "SELECT * FROM collection WHERE installation_unfinished=:installation_unfinished ",
            {"installation_unfinished": 1},
        ).fetchall():
            solution = self._process_solution_row(dict(row), close=False)
            solutions_list.append(solution)

        if close:
            self.close_current_connection()

        return solutions_list

    def update_solution(
        self,
        catalog_id,
        coordinates: ICoordinates,
        solution_attrs: dict,
        supported_attrs: list,
        close=True,
    ):
        exec_str = "UPDATE collection SET last_execution=:cur_date"
        exec_args = {
            "cur_date": datetime.now().isoformat(),
            "catalog_id": catalog_id,
            "group": coordinates.group(),
            "name": coordinates.name(),
            "version": coordinates.version(),
        }

        for key in supported_attrs:
            if key in solution_attrs:
                col = self._as_db_col(key)
                exec_str += f", {col}=:{key}"
                exec_args[key] = get_dict_entry(solution_attrs, key)

        exec_str += ' WHERE catalog_id=:catalog_id AND "group"=:group AND name=:name AND version=:version'

        cursor = self.get_cursor()
        cursor.execute(exec_str, exec_args)

        if close:
            self.close_current_connection()

    def add_or_replace_solution(
        self, catalog_id, coordinates: ICoordinates, solution_attrs, close=True
    ):
        solution = self.get_solution_by_catalog_grp_name_version(
            catalog_id, coordinates, close=False
        )
        if solution:
            self.remove_solution(catalog_id, coordinates, close=False)
        self.insert_solution(catalog_id, solution_attrs, close=close)

    def remove_solution(self, catalog_id, coordinates: ICoordinates, close=True):
        solution = self.get_solution_by_catalog_grp_name_version(
            catalog_id, coordinates, close=False
        )
        if not solution:
            return
        collection_id = solution.internal()["collection_id"]

        cursor = self.get_cursor()
        cursor.execute(
            "DELETE FROM collection_tag WHERE collection_id=:collection_id",
            {"collection_id": collection_id},
        )

        cursor.execute(
            "DELETE FROM collection_author WHERE collection_id=:collection_id",
            {"collection_id": collection_id},
        )

        cursor.execute(
            "DELETE FROM collection_citation WHERE collection_id=:collection_id",
            {"collection_id": collection_id},
        )

        cursor.execute(
            "DELETE FROM collection_argument WHERE collection_id=:collection_id",
            {"collection_id": collection_id},
        )

        cursor.execute(
            "DELETE FROM collection_custom WHERE collection_id=:collection_id",
            {"collection_id": collection_id},
        )

        cursor.execute(
            "DELETE FROM collection_collection WHERE collection_id_child=:collection_id",
            {"collection_id": collection_id},
        )

        cursor.execute(
            "DELETE FROM cover WHERE collection_id=:collection_id",
            {"collection_id": collection_id},
        )

        cursor.execute(
            "DELETE FROM documentation WHERE collection_id=:collection_id",
            {"collection_id": collection_id},
        )

        cursor.execute(
            "DELETE FROM tag "
            "WHERE NOT EXISTS (SELECT st.tag_id FROM collection_tag st "
            "WHERE tag.tag_id = st.tag_id)"
        )

        cursor.execute(
            "DELETE FROM argument "
            "WHERE NOT EXISTS (SELECT sa.argument_id FROM collection_argument sa "
            "WHERE argument.argument_id = sa.argument_id)"
        )

        cursor.execute(
            "DELETE FROM citation "
            "WHERE NOT EXISTS (SELECT sc.collection_citation_id FROM collection_citation sc "
            "WHERE citation.citation_id = sc.citation_id)"
        )

        cursor.execute(
            "DELETE FROM author "
            "WHERE NOT EXISTS (SELECT sa.collection_author_id FROM collection_author sa "
            "WHERE author.author_id = sa.author_id)"
        )

        cursor.execute(
            "DELETE FROM custom "
            "WHERE NOT EXISTS (SELECT sa.custom_id FROM collection_custom sa "
            "WHERE custom.custom_id = sa.custom_id)"
        )

        # finally delete solution
        cursor.execute(
            "DELETE FROM collection "
            'WHERE catalog_id=:catalog_id AND "group"=:group AND name=:name AND version=:version',
            {
                "catalog_id": catalog_id,
                "group": coordinates.group(),
                "name": coordinates.name(),
                "version": coordinates.version(),
            },
        )

        if close:
            self.close_current_connection()

    def is_installed(self, catalog_id, coordinates: ICoordinates, close=True):
        r = self.get_solution_by_catalog_grp_name_version(
            catalog_id, coordinates, close=close
        )
        if not r:
            raise LookupError(f"Solution {catalog_id}:{coordinates} not found!")
        return True if r.internal()["installed"] else False

    def __len__(self, close=True):
        cursor = self.get_cursor()

        r = cursor.execute("SELECT COUNT(*) FROM collection").fetchone()
        r = r[0]

        if close:
            self.close_current_connection()

        return r

    @staticmethod
    def _as_db_col(key):
        if key == "group":
            return '"group"'
        return key

    @staticmethod
    def get_collection_column_keys():
        res = CatalogIndex.get_solution_column_keys()
        res.append("installed")
        res.append("installation_unfinished")
        res.append("last_execution")
        res.append("install_date")
        res.append("catalog_id")
        return res



from abc import ABCMeta, abstractmethod
from typing import Optional, List, Dict

from album.core.api.model.catalog import ICatalog
from album.core.api.model.catalog_updates import ICatalogUpdates


class ICatalogHandler:
    __metaclass__ = ABCMeta

    @abstractmethod
    def create_cache_catalog(self):
        """Creates the local catalog on the disk from the available initial catalogs.

        Does not contain a DB file. Used only when album starts the first time.

        """
        raise NotImplementedError

    @abstractmethod
    def add_initial_catalogs(self):
        """Adds the initial catalogs to the catalog_collection.

        Copies/downloads them from their src to their local cache. (Except local_catalog)

        """
        raise NotImplementedError

    @abstractmethod
    def add_by_src(self, source, branch_name="main") -> ICatalog:
        """Adds a catalog. Creates them from their src. (Git, network-drive, folder outside cache, etc.)"""
        raise NotImplementedError

    @abstractmethod
    def _add_to_index(self, catalog: ICatalog) -> int:
        """Adds a catalog to the collection index.

        Args:
            catalog: The catalog object

        Returns:
            The database ID of the catalog.

        """
        raise NotImplementedError

    @abstractmethod
    def get_by_id(self, catalog_id) -> ICatalog:
        """Looks up a catalog by its id and returns it."""
        raise NotImplementedError

    @abstractmethod
    def get_by_src(self, src) -> ICatalog:
        """Returns the catalog object of a given url if configured."""
        raise NotImplementedError

    @abstractmethod
    def get_by_name(self, name) -> ICatalog:
        """Looks up a catalog by its id and returns it."""
        raise NotImplementedError

    @abstractmethod
    def get_by_path(self, path) -> ICatalog:
        """Looks up a catalog by its id and returns it."""
        raise NotImplementedError

    @abstractmethod
    def get_all(self) -> List[ICatalog]:
        """Creates the catalog objects from the catalogs specified in the configuration."""
        raise NotImplementedError

    @abstractmethod
    def get_cache_catalog(self) -> ICatalog:
        """Returns the first cache catalog in the configuration (Reads db table from top)."""
        raise NotImplementedError

    @abstractmethod
    def create_new_metadata(self, local_path, name, catalog_type):
        """Creates a new catalog on the disk."""
        raise NotImplementedError

    @abstractmethod
    def update_by_name(self, catalog_name) -> bool:
        """Updates a catalog by its name."""
        raise NotImplementedError

    @abstractmethod
    def update_all(self) -> List[bool]:
        """Updates all available catalogs"""
        raise NotImplementedError

    @abstractmethod
    def update_any(self, catalog_name=None):
        """Updates either all catalogs or one by its name."""
        raise NotImplementedError

    @abstractmethod
    def update_collection(
        self, catalog_name=None, dry_run: bool = False, override: bool = False
    ) -> Dict[str, ICatalogUpdates]:
        """Includes all new changes from a given catalog (or all catalogs) in the catalog_collection."""
        raise NotImplementedError

    @abstractmethod
    def remove_from_collection_by_path(self, path) -> Optional[ICatalog]:
        """Removes a catalog given by its path from the catalog_collection.

        Thereby deleting all its entries from the collection.

        """
        raise NotImplementedError

    @abstractmethod
    def remove_from_collection_by_name(self, name) -> Optional[ICatalog]:
        """Removes a catalog given its name from the catalog_collection."""
        raise NotImplementedError

    @abstractmethod
    def remove_from_collection_by_src(self, src) -> Optional[ICatalog]:
        """Removes a catalog given its src from the catalog_collection."""
        raise NotImplementedError

    @abstractmethod
    def get_installed_solutions(self, catalog: ICatalog) -> list:
        """Get all installed solutions of a catalog"""
        raise NotImplementedError

    @abstractmethod
    def get_all_as_dict(self) -> dict:
        """Get all catalogs as dictionary."""
        raise NotImplementedError

    @abstractmethod
    def set_version(self, catalog: ICatalog):
        raise NotImplementedError



import os
import tempfile
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Generator

from album.ci.utils.zenodo_api import ZenodoMetadata
from album.runner import album_logging
from git import Repo

from album.api import Album
from album.ci.controller.zenodo_manager import ZenodoManager
from album.ci.utils.continuous_integration import get_ssh_url, create_report
from album.core.model.catalog import Catalog, retrieve_index_files_from_src
from album.core.model.default_values import DefaultValues
from album.core.utils.export.changelog import get_changelog_file_name
from album.core.utils.operations import view_operations
from album.core.utils.operations.file_operations import (
    get_dict_from_yml,
    write_dict_to_yml,
    get_dict_entry,
    copy,
    force_remove,
    zip_folder,
)
from album.core.utils.operations.git_operations import (
    checkout_branch,
    add_files_commit_and_push,
    retrieve_files_from_head_last_commit,
    configure_git,
    add_tag,
    retrieve_files_from_head,
)
from album.core.utils.operations.resolve_operations import dict_to_coordinates, as_tag

module_logger = album_logging.get_active_logger


class ReleaseManager:
    """Class for handling a catalog as administrator."""

    configuration = None

    def __init__(
        self,
        album_instance: Album,
        catalog_name,
        catalog_path,
        catalog_src,
        force_retrieve,
    ):
        self.catalog_name = catalog_name
        self.catalog_path = catalog_path
        self.catalog_src = catalog_src

        self.configuration = album_instance.configuration()
        if not self.configuration.is_setup():
            self.configuration.setup()

        self.catalog = Catalog(
            None, name=self.catalog_name, path=catalog_path, src=self.catalog_src
        )
        self.force_retrieve = force_retrieve
        self.album_instance = album_instance

    def _open_repo(self) -> Generator[Repo, None, None]:
        repo = self.catalog.retrieve_catalog(
            force_retrieve=self.force_retrieve, update=False
        )

        return repo

    def close(self):
        if self.catalog:
            self.catalog.dispose()

    def configure_repo(self, user_name, user_email):
        module_logger().info(
            "Configuring repository using:\n\tusername:\t%s\n\temail:\t%s"
            % (user_name, user_email)
        )
        with self._open_repo() as repo:
            configure_git(repo, user_email, user_name)

    def configure_ssh(self, project_path):
        module_logger().info("Configure ssh protocol for repository %s" % project_path)
        with self._open_repo() as repo:
            if not repo.remote().url.startswith("git"):
                repo.remote().set_url(get_ssh_url(project_path, self.catalog_src))

    @staticmethod
    def _get_yml_dict(head):
        yml_file_path = retrieve_files_from_head_last_commit(
            head, DefaultValues.solution_yml_default_name.value
        )[0]
        yml_dict = get_dict_from_yml(yml_file_path)

        return [yml_dict, yml_file_path]

    def zenodo_publish(self, branch_name, zenodo_base_url, zenodo_access_token):
        zenodo_base_url, zenodo_access_token = self._prepare_zenodo_arguments(
            zenodo_base_url, zenodo_access_token
        )

        zenodo_manager = self._get_zenodo_manager(zenodo_access_token, zenodo_base_url)

        with self._open_repo() as repo:
            head = checkout_branch(repo, branch_name)

            # get the yml file to release from current branch
            yml_dict, yml_file_path = self._get_yml_dict(head)

            with tempfile.TemporaryDirectory(
                dir=self.album_instance.configuration().tmp_path()
            ) as tmp:

                # checkout files
                files, file_names = self._get_release_files(repo, yml_dict, tmp)

                # query deposit
                deposit_name = self._get_deposit_metadata(yml_dict).title
                deposit_id = get_dict_entry(yml_dict, "deposit_id", allow_none=False)

                module_logger().info(
                    "Get unpublished deposit with deposit id %s..." % deposit_id
                )
                deposit = zenodo_manager.zenodo_get_unpublished_deposit_by_id(
                    deposit_id, deposit_name, expected_files=file_names
                )

                # publish to zenodo
                deposit.publish()

        add_tag(repo, as_tag(dict_to_coordinates(yml_dict)))

        module_logger().info(
            "Published unpublished deposit with deposit id %s..." % deposit_id
        )

    def _get_release_files(self, repo, yml_dict, tmp):
        coordinates = dict_to_coordinates(yml_dict)
        solution_relative_path = (
            self.configuration.get_solution_path_suffix_unversioned(coordinates)
        )
        solution_dir_in_repo = Path(repo.working_tree_dir).joinpath(
            solution_relative_path
        )
        zip = Path(tmp).joinpath(DefaultValues.solution_zip_default_name.value)
        zip_folder(solution_dir_in_repo, zip)
        docker_name = solution_dir_in_repo.joinpath("Dockerfile")
        changelog_name = solution_dir_in_repo.joinpath(get_changelog_file_name())
        documentation_paths = self._get_documentation_paths(
            solution_dir_in_repo, yml_dict
        )
        cover_paths = self._get_cover_paths(solution_dir_in_repo, yml_dict)
        solution_yml = solution_dir_in_repo.joinpath(
            DefaultValues.solution_yml_default_name.value
        )
        files = [str(solution_yml), str(zip), str(docker_name)]
        for doc in documentation_paths:
            files.append(str(doc))
        for cover in cover_paths:
            files.append(str(cover))
        if changelog_name.exists():
            files.append(str(changelog_name))
        return files, [Path(file).name for file in files]

    def zenodo_upload(
        self, branch_name, zenodo_base_url, zenodo_access_token, report_file
    ):
        zenodo_base_url, zenodo_access_token = self._prepare_zenodo_arguments(
            zenodo_base_url, zenodo_access_token
        )

        zenodo_manager = self._get_zenodo_manager(zenodo_access_token, zenodo_base_url)
        with self._open_repo() as repo:
            head = checkout_branch(repo, branch_name)

            # get the yml file to release
            yml_dict, yml_file_path = self._get_yml_dict(head)

            # get metadata from yml_file
            try:
                deposit_id = yml_dict["deposit_id"]
            except KeyError:
                deposit_id = None
            coordinates = dict_to_coordinates(yml_dict)

            # do not upload SNAPSHOT versions
            if coordinates.version().endswith("SNAPSHOT"):
                module_logger().info(
                    "Will not upload SNAPSHOT version to zenodo! Skipping..."
                )
                return

            with tempfile.TemporaryDirectory(
                dir=self.album_instance.configuration().tmp_path()
            ) as tmp:

                files, file_names = self._get_release_files(repo, yml_dict, tmp)

                # get the release deposit. Either a new one or an existing one to perform an update on
                deposit = zenodo_manager.zenodo_get_deposit(
                    self._get_deposit_metadata(yml_dict), deposit_id
                )
                module_logger().info(
                    "Deposit %s successfully retrieved..." % deposit.id
                )

                # include doi and ID in yml
                yml_dict["doi"] = deposit.metadata.prereserve_doi["doi"]
                yml_dict["deposit_id"] = deposit.id
                write_dict_to_yml(yml_file_path, yml_dict)

                if deposit.files:
                    for file in deposit.files:
                        if file not in file_names:
                            # remove file from deposit
                            zenodo_manager.zenodo_delete(deposit, file)

                for file in files:
                    deposit = zenodo_manager.zenodo_upload(deposit, file)

        module_logger().info("Deposit %s successfully retrieved..." % deposit_id)

        if report_file:
            report_file = create_report(
                report_file,
                {"DOI": yml_dict["doi"], "DEPOSIT_ID": yml_dict["deposit_id"]},
            )
            module_logger().info("Created report file under %s" % str(report_file))

    @staticmethod
    def _prepare_zenodo_arguments(zenodo_base_url: str, zenodo_access_token: str):
        if zenodo_base_url is None or zenodo_access_token is None:
            raise RuntimeError(
                "Zenodo base URL or Zenodo access token invalid! "
                "See https://zenodo.org/ documentation for information!"
            )

        zenodo_base_url = zenodo_base_url.strip()
        zenodo_access_token = zenodo_access_token.strip()

        if zenodo_base_url == "" or zenodo_access_token == "":
            raise RuntimeError("Empty zenodo base_url or zenodo_access_token!")

        module_logger().info("Using zenodo URL: %s" % zenodo_base_url)

        return zenodo_base_url, zenodo_access_token

    def update_index(self, branch_name, doi, deposit_id):
        with self._open_repo() as repo:
            head = checkout_branch(repo, branch_name)

            # always use remote index
            with TemporaryDirectory(dir=self.configuration.tmp_path()) as tmp_dir:
                repo = Path(tmp_dir).joinpath("repo")
                try:
                    index_db, index_meta = retrieve_index_files_from_src(
                        self.catalog_src,
                        branch_name=self.catalog.branch_name(),
                        tmp_dir=repo,
                    )
                    if index_db.exists():
                        copy(index_db, self.catalog.index_file_path())
                    else:
                        module_logger().warning(
                            "Index not downloadable! Using Index in merge request branch!"
                        )
                finally:
                    force_remove(repo)

            self.album_instance.load_catalog_index(self.catalog)

            yml_dict, yml_file_path = self._get_yml_dict(head)

            # update doi and deposit_id if given
            log_message_doi = "without a doi"
            if doi:
                yml_dict["doi"] = doi
                log_message_doi = "with doi %s" % doi

            log_message_deposit = "without deposit"
            if deposit_id:
                yml_dict["deposit_id"] = deposit_id
                log_message_deposit = "in zenodo deposit %s" % deposit_id

            if doi or deposit_id:
                write_dict_to_yml(yml_file_path, yml_dict)

        module_logger().info(
            "Update index to include solution from branch %s %s %s!"
            % (branch_name, log_message_doi, log_message_deposit)
        )

        self.catalog.index().update(dict_to_coordinates(yml_dict), yml_dict)
        self.catalog.index().save()
        self.catalog.index().export(self.catalog.solution_list_path())

    def commit_changes(self, branch_name, ci_user_name, ci_user_email):
        with self._open_repo() as repo:
            head = checkout_branch(repo, branch_name)

            yml_dict, yml_file = self._get_yml_dict(head)

            # get files to release
            coordinates = dict_to_coordinates(yml_dict)
            solution_relative_path = (
                self.configuration.get_solution_path_suffix_unversioned(coordinates)
            )
            files = retrieve_files_from_head(
                head, str(solution_relative_path) + os.sep + "*", number_of_files=-1
            )

            commit_files = [yml_file] + files
            if not all([Path(f).is_file() for f in commit_files]):
                raise FileNotFoundError(
                    "Invalid deploy request or broken catalog repository!"
                )

            commit_msg = 'Prepared branch "%s" for merging.' % branch_name

            add_files_commit_and_push(
                head,
                commit_files,
                commit_msg,
                push=False,
                username=ci_user_name,
                email=ci_user_email,
            )

        return True

    def merge(self, branch_name, dry_run, push_option, ci_user_name, ci_user_email):
        with self._open_repo() as repo:
            head = checkout_branch(repo, branch_name)

            commit_files = [
                self.catalog.solution_list_path(),
                self.catalog.index_file_path(),
            ]
            if not all([Path(f).is_file() for f in commit_files]):
                raise FileNotFoundError(
                    "Invalid deploy request or broken catalog repository!"
                )

            commit_msg = "Updated index."

            module_logger().info("Prepare merging...")

            add_files_commit_and_push(
                head,
                commit_files,
                commit_msg,
                push=not dry_run,
                push_option_list=push_option,
                username=ci_user_name,
                email=ci_user_email,
            )

    def _get_zenodo_manager(self, zenodo_access_token, zenodo_base_url):
        # TODO in case one wants to reuse the zenodo manager, this needs to be smarter, but be aware that another call
        #  to this method might have different parameters
        return ZenodoManager(zenodo_base_url, zenodo_access_token)

    @staticmethod
    def _get_documentation_paths(base_dir, yml_dict: dict):
        documentation_paths = []
        if "documentation" in yml_dict.keys():
            documentation_list = yml_dict["documentation"]
            if not isinstance(documentation_list, list):
                documentation_list = [documentation_list]

            for documentation_name in documentation_list:
                documentation_paths.append(base_dir.joinpath(documentation_name))
        return documentation_paths

    @staticmethod
    def _get_cover_paths(base_dir, yml_dict: dict):
        cover_paths = []
        if "covers" in yml_dict.keys():
            cover_list = yml_dict["covers"]
            if not isinstance(cover_list, list):
                cover_list = [cover_list]

            for cover in cover_list:
                if "source" in cover:
                    cover_paths.append(base_dir.joinpath(cover["source"]))
        return cover_paths

    def _get_deposit_metadata(self, solution_meta):
        if "title" in solution_meta:
            deposit_name = solution_meta["title"]
        else:
            deposit_name = str(dict_to_coordinates(solution_meta))
        if "solution_creators" in solution_meta:
            authors = solution_meta["solution_creators"]
        else:
            authors = []
        creators = [{"name": author} for author in authors]
        if not creators:
            creators = [{"name": "unknown"}]
        description = (
            "Album solution - for more information read https://album.solutions."
        )
        if "description" in solution_meta:
            description = solution_meta["description"]
        license = None
        if "license" in solution_meta:
            license = solution_meta["license"]
        version = None
        if "version" in solution_meta:
            version = solution_meta["version"]
        related_identifiers = [
            {"relation": "cites", "identifier": DefaultValues.album_cite_doi.value}
        ]
        references = [
            view_operations.get_citation_as_string(
                {
                    "doi": DefaultValues.album_cite_doi.value,
                    "url": DefaultValues.album_cite_url.value,
                    "text": DefaultValues.album_cite_text.value,
                }
            )
        ]
        if "cite" in solution_meta:
            for citation in solution_meta["cite"]:
                if "doi" in citation:
                    related_identifiers.append(
                        {"relation": "cites", "identifier": citation["doi"]}
                    )
                if "text" in citation:
                    references.append(view_operations.get_citation_as_string(citation))

        return ZenodoMetadata.default_values(
            deposit_name,
            creators,
            description,
            license,
            version,
            related_identifiers,
            references,
        )



from abc import ABCMeta, abstractmethod

from album.core.api.controller.collection.catalog_handler import ICatalogHandler
from album.core.api.controller.collection.solution_handler import ISolutionHandler
from album.core.api.model.catalog import ICatalog
from album.core.api.model.collection_index import ICollectionIndex
from album.core.api.model.collection_solution import ICollectionSolution
from album.runner.core.api.model.coordinates import ICoordinates


class ICollectionManager:
    """The Album Catalog Collection class.

    An album framework installation instance can hold arbitrarily many catalogs. This class holds all configured
    catalogs in memory and is mainly responsible to resolve (look up) solutions in all these catalogs.
    It is not responsible for resolving local paths and files or remembering what is already installed!
    Please use the resolve manager for this!
    Additionally, catalogs can be configured via this class.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def load_or_create(self):
        raise NotImplementedError

    @abstractmethod
    def catalogs(self) -> ICatalogHandler:
        raise NotImplementedError

    @abstractmethod
    def solutions(self) -> ISolutionHandler:
        raise NotImplementedError

    @abstractmethod
    def get_index_as_dict(self):
        raise NotImplementedError

    @abstractmethod
    def resolve_installed(self, resolve_solution) -> ICollectionSolution:
        """Resolves an input. Expects solution to be installed.

        Args:
            resolve_solution:
                What to resolve. Either path, doi, group:name:version, catalog:group:name:version, url


        Returns:
            The resolve result, including the loaded solution.
        """
        raise NotImplementedError

    @abstractmethod
    def resolve_installed_and_load(self, resolve_solution) -> ICollectionSolution:
        """Resolves an input. Expects solution to be installed.

        Args:
            resolve_solution:
                What to resolve. Either path, doi, group:name:version, catalog:group:name:version, url


        Returns:
            The resolve result, including the loaded solution.
        """
        raise NotImplementedError

    @abstractmethod
    def resolve_and_load(self, resolve_solution) -> ICollectionSolution:
        """Resolves a string input and loads its content.

        Downloads a catalog if not already cached.

        Args:
            resolve_solution:
                What to resolve. Either path, doi, group:name:version, catalog:group:name:version, url

        Returns:
            The resolve result, including the loaded solution.

        """
        raise NotImplementedError

    @abstractmethod
    def resolve_and_load_catalog_coordinates(
        self, catalog: ICatalog, coordinates: ICoordinates
    ) -> ICollectionSolution:
        """Resolves a string input and loads its content.

        Downloads a catalog if not already cached.

        Args:
            catalog:
                Catalog to resolve in.
            coordinates:
                Coordinates to resolve.

        Returns:
            The resolve result, including the loaded solution.

        """
        raise NotImplementedError

    @abstractmethod
    def resolve_and_load_coordinates(
        self, coordinates: ICoordinates
    ) -> ICollectionSolution:
        """Resolves a string input and loads its content.

        Downloads a catalog if not already cached.

        Args:
            coordinates:
                Coordinates to resolve.

        Returns:
            The resolve result, including the loaded solution.

        """
        raise NotImplementedError

    @abstractmethod
    def resolve(self, resolve_solution) -> ICollectionSolution:
        """Resolves a string input and loads its content.

        Downloads a catalog if not already cached.

        Args:
            resolve_solution:
                What to resolve. Either path, doi, group:name:version, catalog:group:name:version, url

        Returns:
            The resolve result

        """
        raise NotImplementedError

    @abstractmethod
    def resolve_parent(self, parent_dict: dict) -> ICollectionSolution:
        raise NotImplementedError

    def get_collection_index(self) -> ICollectionIndex:
        raise NotImplementedError

    @abstractmethod
    def retrieve_and_load_resolve_result(self, resolve_result: ICollectionSolution):
        raise NotImplementedError



from typing import List, Optional

from album.core.api.model.catalog import ICatalog
from album.core.api.model.catalog_updates import (
    ISolutionChange,
    ChangeType,
    ICatalogAttributeChange,
    ICatalogUpdates,
)
from album.runner.core.api.model.coordinates import ICoordinates


class SolutionChange(ISolutionChange):
    def __init__(
        self,
        coordinates: ICoordinates,
        change_type: ChangeType,
        change_log: Optional[str] = None,
        solution_status: Optional[dict] = None,
    ):
        self._coordinates = coordinates
        self._change_type = change_type
        self._change_log = change_log
        self._solution_status = solution_status

    def __eq__(self, other):
        return (
            isinstance(other, ISolutionChange)
            and other.coordinates() == self._coordinates
            and other.change_type() == self._change_type
            and other.change_log() == self._change_log
            and other.solution_status() == self._solution_status
        )

    def as_dict(self):
        return {
            "group": self._coordinates.group(),
            "name": self._coordinates.name(),
            "version": self._coordinates.version(),
            "change_type": self._change_type.name,
            "change_log": self._change_log,
            # solution status only internal
        }

    def coordinates(self) -> ICoordinates:
        return self._coordinates

    def change_type(self) -> ChangeType:
        return self._change_type

    def change_log(self) -> str:
        return self._change_log

    def solution_status(self) -> dict:
        return self._solution_status


class CatalogAttributeChange(ICatalogAttributeChange):
    def __init__(self, attribute: str, old_value, new_value):
        self._attribute = attribute
        self._old_value = old_value
        self._new_value = new_value

    def old_value(self):
        return self._old_value

    def new_value(self):
        return self._new_value

    def attribute(self) -> str:
        return self._attribute


class CatalogUpdates(ICatalogUpdates):
    _catalog: ICatalog
    _catalog_attribute_changes: List[ICatalogAttributeChange] = []
    _solution_changes: List[ISolutionChange] = []

    def __init__(
        self,
        catalog: ICatalog,
        solution_changes: Optional[List[ISolutionChange]] = None,
        catalog_attribute_changes: Optional[List[ICatalogAttributeChange]] = None,
    ) -> None:
        if catalog_attribute_changes is None:
            catalog_attribute_changes = []
        if solution_changes is None:
            solution_changes = []
        self._catalog = catalog
        self._solution_changes = solution_changes
        self._catalog_attribute_changes = catalog_attribute_changes

    def as_dict(self):
        solution_changes_as_dict = []
        for change in self._solution_changes:
            solution_changes_as_dict.append(change.as_dict())
        return {
            "catalog": {"name": self._catalog.name()},
            "solution_changes": solution_changes_as_dict,
        }

    def __str__(self) -> str:
        return str(self.as_dict())

    def catalog(self) -> ICatalog:
        return self._catalog

    def catalog_attribute_changes(self) -> List[ICatalogAttributeChange]:
        return self._catalog_attribute_changes

    def solution_changes(self):
        return self._solution_changes



import sys

from album.api import Album
from album.core.utils.operations.solution_operations import (
    get_deploy_dict,
    serialize_json,
)
from album.core.utils.operations.view_operations import (
    get_solution_as_string,
    get_updates_as_string,
    get_index_as_string,
    get_search_result_as_string,
)
from album.runner.album_logging import get_active_logger
from album.runner.core.api.model.solution import ISolution
from album.runner.core.model.script_creator import ScriptCreator

module_logger = get_active_logger


def add_catalog(album_instance: Album, args) -> None:
    album_instance.add_catalog(args.src)


def remove_catalog(album_instance: Album, args) -> None:
    album_instance.remove_catalog_by_name(args.name)


def update(album_instance: Album, args):
    album_instance.update(getattr(args, "catalog", None))


def upgrade(album_instance: Album, args):
    updates = album_instance.upgrade(
        getattr(args, "catalog", None), dry_run=args.dry_run, override=args.override
    )
    print_json = _get_print_json(args)
    if print_json:
        print(_as_json(updates))
    else:
        res = ""
        if args.dry_run:
            res += "An upgrade would apply the following updates:\n"
        else:
            res += "Applied the following updates:\n"
        res += get_updates_as_string(updates)
        module_logger().info(res)


def deploy(album_instance: Album, args):
    album_instance.deploy(
        args.path,
        args.catalog,
        args.dry_run,
        args.push_option,
        args.git_email,
        args.git_name,
        args.force_deploy,
        args.changelog,
    )


def undeploy(album_instance: Album, args):
    album_instance.undeploy(
        args.path,
        args.catalog,
        args.dry_run,
        args.push_option,
        args.git_email,
        args.git_name,
    )


def install(album_instance: Album, args):
    album_instance.install(str(args.path), sys.argv)


def uninstall(album_instance: Album, args):
    album_instance.uninstall(str(args.path), args.uninstall_deps, sys.argv)


def info(album_instance: Album, args):
    solution_path = args.path
    resolve_result = album_instance.resolve(str(args.path))
    print_json = _get_print_json(args)
    solution = resolve_result.loaded_solution()
    if print_json:
        deploy_dict = get_deploy_dict(solution)
        print(_as_json(deploy_dict))
    else:
        res = get_solution_as_string(solution, solution_path)
        module_logger().info(res)


def run(album_instance: Album, args):
    album_instance.run(str(args.path), argv=sys.argv)


def search(album_instance: Album, args):
    print_json = _get_print_json(args)
    search_result = album_instance.search(args.keywords)
    if print_json:
        print(_as_json(search_result))
    else:
        res = get_search_result_as_string(args, search_result)
        module_logger().info(res)


def test(album_instance: Album, args):
    album_instance.test(str(args.path), sys.argv)


def clone(album_instance: Album, args):
    album_instance.clone(args.src, args.target_dir, args.name)


def index(album_instance: Album, args):
    index_dict = album_instance.get_index_as_dict()
    print_json = _get_print_json(args)
    if print_json:
        print(_as_json(index_dict))
    else:
        res = get_index_as_string(index_dict)
        module_logger().info(res)


def repl(album_instance: Album, args):
    """Function corresponding to the `repl` subcommand of `album`."""
    # resolve the input
    resolve_result = album_instance.resolve_installed(str(args.path))
    album_instance.run_solution_script(resolve_result, ScriptRepl())
    module_logger().info(
        'Ran REPL for "%s"!' % resolve_result.loaded_solution().coordinates().name()
    )


def _get_print_json(args):
    return getattr(args, "json", False)


def _as_json(data):
    return serialize_json(data)


class ScriptRepl(ScriptCreator):
    def get_execution_block(self, solution_object: ISolution):
        return """
from code import InteractiveConsole
try:
    import readline
except ImportError:
    print("Module readline not available.")
else:
    import rlcompleter
    readline.parse_and_bind("tab: complete")
console = InteractiveConsole(locals={**globals(), **locals()})
console.interact()
"""

    def __init__(self, pop_solution: bool = False, execution_callback=None):
        super().__init__(append_arguments=False)
        self.pop_solution = pop_solution

        if execution_callback is not None and callable(execution_callback):
            self.execution_callback = execution_callback
        else:
            self.reset_callback()

    def reset_callback(self):
        self.execution_callback = lambda: ""



from abc import ABCMeta, abstractmethod

from album.core.api.model.catalog import ICatalog
from album.core.api.model.catalog_updates import ISolutionChange
from album.core.api.model.collection_index import ICollectionIndex
from album.runner.core.api.model.coordinates import ICoordinates
from album.runner.core.api.model.solution import ISolution


class ISolutionHandler:
    __metaclass__ = ABCMeta

    @abstractmethod
    def add_to_cache_catalog(self, active_solution: ISolution, path):
        """Force adds the installation to the local catalog to be cached for running"""
        raise NotImplementedError

    @abstractmethod
    def set_uninstalled(self, catalog: ICatalog, coordinates: ICoordinates):
        raise NotImplementedError

    @abstractmethod
    def set_parent(
        self,
        parent: ICollectionIndex.ICollectionSolution,
        child: ICollectionIndex.ICollectionSolution,
    ):
        raise NotImplementedError

    @abstractmethod
    def remove_parent(self, catalog: ICatalog, coordinates: ICoordinates):
        raise NotImplementedError

    @abstractmethod
    def remove_solution(self, catalog: ICatalog, coordinates: ICoordinates):
        raise NotImplementedError

    @abstractmethod
    def update_solution(
        self, catalog: ICatalog, coordinates: ICoordinates, attrs: dict
    ):
        raise NotImplementedError

    @abstractmethod
    def apply_change(self, catalog: ICatalog, change: ISolutionChange, override: bool):
        raise NotImplementedError

    @abstractmethod
    def set_installed(self, catalog: ICatalog, coordinates: ICoordinates):
        raise NotImplementedError

    @abstractmethod
    def is_installed(self, catalog: ICatalog, coordinates: ICoordinates) -> bool:
        raise NotImplementedError

    @abstractmethod
    def get_solution_package_path(self, catalog: ICatalog, coordinates: ICoordinates):
        """Gets the cache path of a solution in a catalog given its group, name and version.

        Args:
            catalog:
                The catalog this solution lives in.
            coordinates:
                The group affiliation, name, and version of the solution.

        Returns:
            The absolute path to the cache folder of the solution.

        """
        raise NotImplementedError

    @abstractmethod
    def get_solution_installation_path(
        self, catalog: ICatalog, coordinates: ICoordinates
    ):
        """Gets the path of files belonging to an installed solution in a catalog given its group, name and version.

        Args:
            catalog:
                The catalog this solution lives in.
            coordinates:
                The group affiliation, name, and version of the solution.

        Returns:
            The absolute path to the installation folder of the solution.

        """
        raise NotImplementedError

    @abstractmethod
    def get_solution_file(self, catalog: ICatalog, coordinates: ICoordinates):
        """Gets the file to the solution.py file of the extracted solution.zip living inside the catalog.

        Args:
            catalog:
                The catalog this solution belongs to.
            coordinates:
                The group affiliation, name, and version of the solution.

        Returns:
            The path where the file is supposed to be once the zip is extracted during installation.

        """
        raise NotImplementedError

    @abstractmethod
    def retrieve_solution(self, catalog: ICatalog, coordinates: ICoordinates):
        """Downloads or copies a solution from the catalog to the local resource (cache path of the catalog).

        Args:
            catalog:
                The catalog this solution belongs to.
            coordinates:
                The group affiliation, name, and version of the solution.
        Returns:
            The absolute path of the downloaded solution.

        """
        raise NotImplementedError

    @abstractmethod
    def get_solution_zip_suffix(self, coordinates: ICoordinates):
        """Gets the cache zip suffix of a solution given its group, name and version living inside the catalog.

        Args:
            coordinates:
                The group affiliation, name, and version of the solution.

        Returns:
            The the cache zip suffix.

        """
        raise NotImplementedError

    @abstractmethod
    def set_cache_paths(self, solution: ISolution, catalog: ICatalog):
        """Sets the available cache paths of the solution object, given its catalog_name (where it lives)."""
        raise NotImplementedError

    @abstractmethod
    def set_installation_unfinished(self, catalog: ICatalog, coordinates: ICoordinates):
        raise NotImplementedError



import os
from datetime import datetime
from pathlib import Path

from git import Repo

from album.core.api.controller.controller import IAlbumController
from album.core.api.controller.deploy_manager import IDeployManager
from album.core.api.model.catalog import ICatalog
from album.core.model.default_values import DefaultValues
from album.core.utils.export.changelog import (
    create_changelog_file,
    process_changelog_file,
)
from album.core.utils.export.docker import create_docker_file
from album.core.utils.operations.file_operations import (
    copy,
    write_dict_to_yml,
    force_remove,
    folder_empty,
)
from album.core.utils.operations.git_operations import (
    create_new_head,
    add_files_commit_and_push,
    retrieve_default_mr_push_options,
    checkout_main,
    clean_repository,
    remove_files,
    remove_tag,
    add_tag,
    get_tags,
    revert,
)
from album.core.utils.operations.resolve_operations import (
    get_attributes_from_string,
    dict_to_coordinates,
    as_tag_unversioned,
    as_tag,
)
from album.core.utils.operations.solution_operations import get_deploy_dict
from album.runner import album_logging
from album.runner.core.api.model.coordinates import ICoordinates
from album.runner.core.api.model.solution import ISolution

module_logger = album_logging.get_active_logger


class DeployManager(IDeployManager):
    def __init__(self, album: IAlbumController):
        self.album = album

    def deploy(
        self,
        deploy_path,
        catalog_name: str,
        dry_run: bool,
        push_options=None,
        git_email: str = None,
        git_name: str = None,
        force_deploy: bool = False,
        changelog: str = None,
    ):
        if dry_run:
            module_logger().info(
                "Pretending to deploy %s to %s..." % (deploy_path, catalog_name)
            )
        else:
            module_logger().info("Deploying %s to %s..." % (deploy_path, catalog_name))

        deploy_path = Path(deploy_path)
        path_to_solution = self._get_path_to_solution(deploy_path)

        active_solution = self.album.state_manager().load(path_to_solution)
        active_solution.setup().changelog = changelog

        if catalog_name:
            catalog = self.album.catalogs().get_by_name(catalog_name)
        else:
            raise RuntimeError("No catalog specified for deployment!")

        self._deploy(
            catalog,
            active_solution,
            deploy_path,
            dry_run,
            force_deploy,
            push_options,
            git_email,
            git_name,
        )

        if dry_run:
            module_logger().info(
                "Successfully pretended to deploy %s to %s."
                % (deploy_path, catalog_name)
            )
        else:
            module_logger().info(
                "Successfully deployed %s to %s." % (deploy_path, catalog_name)
            )

    def undeploy(
        self,
        solution_to_resolve,
        catalog_name: str,
        dry_run: bool,
        push_options=None,
        git_email: str = None,
        git_name: str = None,
    ):

        exit_msg = (
            "Pretending to remove %s from %s..."
            if dry_run
            else "Removing %s from %s..."
        )
        module_logger().info(exit_msg % (solution_to_resolve, catalog_name))

        coordinates = dict_to_coordinates(
            get_attributes_from_string(solution_to_resolve)
        )
        catalog = self.album.catalogs().get_by_name(catalog_name)
        # check for cache catalog only
        if catalog.is_cache():
            raise RuntimeError(
                "Cannot undeploy from catalog only used for caching! Aborting..."
            )

        with catalog.retrieve_catalog(
            self.get_download_path(catalog), force_retrieve=True
        ) as repo:
            # load index
            catalog.set_index_path(
                Path(repo.working_tree_dir).joinpath(
                    DefaultValues.catalog_index_file_name.value
                )
            )
            self.album.migration_manager().load_index(catalog)

            # get versions from catalog
            ordered_version_tags = self._get_tags_for_coordinates(repo, coordinates)
            current_version_tag = (
                ordered_version_tags[0] if len(ordered_version_tags) > 0 else None
            )
            previous_version_tag = (
                ordered_version_tags[1] if len(ordered_version_tags) > 1 else None
            )

            version_to_be_removed = as_tag(coordinates)
            if not version_to_be_removed in ordered_version_tags:
                raise LookupError(
                    "Cannot find %s in deployed versions %s"
                    % (version_to_be_removed, ordered_version_tags)
                )

            self._remove_from_downloaded_catalog(catalog, coordinates, dry_run)

            if not previous_version_tag:
                # remove files from catalog, update db, remove tag
                self._remove_db_entry_and_files(
                    repo, coordinates, dry_run, push_options, git_email, git_name
                )
            else:
                if current_version_tag == version_to_be_removed:
                    # roll back to last version
                    self._remove_db_entry_and_revert_files(
                        repo,
                        coordinates,
                        previous_version_tag,
                        dry_run,
                        push_options,
                        git_email,
                        git_name,
                    )
                else:
                    # remove tag, update db
                    self._remove_db_entry_and_tag(
                        repo, coordinates, dry_run, push_options, git_email, git_name
                    )

        # refresh the local index of the catalog
        self.album.migration_manager().refresh_index(catalog)

        exit_msg = (
            "Successfully pretended to remove %s from %s."
            if dry_run
            else "Successfully removed %s from %s."
        )
        module_logger().info(exit_msg % (solution_to_resolve, catalog_name))

    def _deploy(
        self,
        catalog: ICatalog,
        active_solution: ISolution,
        deploy_path: Path,
        dry_run: bool,
        force_deploy: bool,
        push_options: list,
        git_email=None,
        git_name=None,
    ):
        # check for cache catalog only
        if catalog.is_cache():
            raise RuntimeError(
                "Cannot deploy to catalog only used for caching! Aborting..."
            )

        dl_path = self.get_download_path(catalog)

        # a catalog is always a repository
        with catalog.retrieve_catalog(dl_path, force_retrieve=True) as repo:
            # load index
            catalog.set_index_path(
                Path(repo.working_tree_dir).joinpath(
                    DefaultValues.catalog_index_file_name.value
                )
            )
            self.album.migration_manager().load_index(catalog)

            # requires a loaded index
            process_changelog_file(catalog, active_solution, deploy_path)

            if catalog.type() == "direct":
                self._deploy_to_direct_catalog(
                    repo,
                    catalog,
                    active_solution,
                    deploy_path,
                    dry_run,
                    force_deploy,
                    push_options,
                    git_email,
                    git_name,
                )
            elif catalog.type() == "request":
                self._deploy_to_request_catalog(
                    repo,
                    catalog,
                    active_solution,
                    deploy_path,
                    dry_run,
                    push_options,
                    git_email,
                    git_name,
                )
            else:
                raise NotImplementedError("type %s not supported!" % catalog.type())

    def _deploy_to_direct_catalog(
        self,
        repo: Repo,
        catalog: ICatalog,
        active_solution: ISolution,
        deploy_path: Path,
        dry_run: bool,
        force_deploy: bool,
        push_options: list,
        git_email=None,
        git_name=None,
    ):
        """Routine to deploy to a direct catalog"""
        # adding solutions in the SQL databse for catalog type "direct" done on user side, hence the timestamp
        active_solution.setup()["timestamp"] = datetime.strftime(
            datetime.now(), "%Y-%m-%dT%H:%M:%S.%f"
        )

        self._add_to_downloaded_catalog(catalog, active_solution, dry_run, force_deploy)

        solution_files = self._deploy_routine_in_local_src(
            catalog, repo, active_solution, deploy_path
        )
        commit_files = solution_files + [catalog.index_file_path()]
        solution_root = str(
            self.album.configuration().get_solution_path_suffix_unversioned(
                active_solution.coordinates()
            )
        )

        # merge and push
        if not dry_run:
            try:
                self._push_directly(
                    active_solution.coordinates(),
                    repo,
                    [solution_root],
                    commit_files,
                    dry_run,
                    push_options,
                    git_email,
                    git_name,
                )
                add_tag(repo, as_tag(active_solution.coordinates()))
            except Exception as e:
                module_logger().error(
                    "Pushing to catalog failed! Rolling back deployment..."
                )
                try:
                    clean_repository(repo)
                    for export in solution_files:
                        export.unlink()
                finally:
                    raise e

            # refresh the local index of the catalog
            self.album.migration_manager().refresh_index(catalog)
        else:
            module_logger().info(
                "Would commit the changes and push to %s..." % catalog.src()
            )
            module_logger().info("Would refresh the index from src")

    def _deploy_to_request_catalog(
        self,
        repo: Repo,
        catalog: ICatalog,
        active_solution: ISolution,
        deploy_path: Path,
        dry_run: bool,
        push_options: list,
        git_email=None,
        git_name=None,
    ):
        """Routine to deploy to a request catalog."""

        # include files/folders in catalog
        mr_files = self._deploy_routine_in_local_src(
            catalog, repo, active_solution, deploy_path
        )

        if not push_options:
            push_options = retrieve_default_mr_push_options(catalog.src())

        self._create_merge_request(
            active_solution.coordinates(),
            repo,
            mr_files,
            dry_run,
            push_options,
            git_email,
            git_name,
        )

    def _deploy_routine_in_local_src(
        self,
        catalog: ICatalog,
        repo: Repo,
        active_solution: ISolution,
        deploy_path: Path,
    ):
        """Performs all routines a deploy process needs to do locally.

        Returns:
            solution zip file and additional attachments.

        """
        solution_files = self._collect_solution_files(
            catalog, repo.working_tree_dir, active_solution, deploy_path
        )

        return solution_files

    def get_download_path(self, catalog: ICatalog):
        return Path(self.album.configuration().cache_path_download()).joinpath(
            catalog.name()
        )

    def _collect_solution_files(
        self,
        catalog: ICatalog,
        catalog_local_src: str,
        active_solution: ISolution,
        deploy_path: Path,
    ):
        coordinates = active_solution.coordinates()

        catalog_solution_local_src_path = Path(catalog_local_src).joinpath(
            self.album.configuration().get_solution_path_suffix_unversioned(coordinates)
        )

        res = []
        if deploy_path.is_file():
            solution_path = catalog_solution_local_src_path.joinpath(
                DefaultValues.solution_default_name.value
            )
            res.append(copy(deploy_path, solution_path))
        else:
            for subdir, dirs, files in os.walk(deploy_path):
                for file in files:
                    filepath = subdir + os.sep + file
                    rel_path = os.path.relpath(filepath, deploy_path)
                    target = catalog_solution_local_src_path.joinpath(rel_path)
                    res.append(copy(filepath, target))

        res.append(create_docker_file(active_solution, catalog_solution_local_src_path))
        res.append(
            self._create_yaml_file_in_local_src(
                active_solution, catalog_solution_local_src_path
            )
        )
        res.append(
            create_changelog_file(
                active_solution, catalog, catalog_solution_local_src_path
            )
        )

        return res

    def _get_tmp_dir(self):
        return self.album.configuration().tmp_path()

    @staticmethod
    def _get_tags_for_coordinates(repo, coordinates: ICoordinates):
        tags = get_tags(repo)
        tag_start = as_tag_unversioned(coordinates)
        return [tag for tag in tags if tag.startswith(tag_start)]

    @staticmethod
    def _add_to_downloaded_catalog(
        catalog: ICatalog, active_solution: ISolution, dry_run: bool, force_deploy: bool
    ):
        """Updates the index in the downloaded repository!"""
        if not dry_run:
            catalog.add(active_solution, force_overwrite=force_deploy)
        else:
            module_logger().info(
                "Would add the solution %s to index..."
                % active_solution.coordinates().name()
            )

    @staticmethod
    def _remove_from_downloaded_catalog(
        catalog: ICatalog, coordinates: ICoordinates, dry_run: bool
    ):
        """Updates the index in the downloaded repository!"""
        if not dry_run:
            catalog.remove(coordinates)
        else:
            module_logger().info(
                "Would remove the solution %s from index..." % coordinates.name()
            )

    @staticmethod
    def _clear_deploy_target_path(target_path: Path, force_deploy: bool):
        """Clears the target path (locally) where the solution is supposed to be deployed to."""
        if target_path.is_dir() and not folder_empty(target_path):
            if force_deploy:
                force_remove(target_path)
            else:
                raise RuntimeError(
                    "The deploy target folder is not empty! Enable --force-deploy to continue!"
                )

    @staticmethod
    def _get_path_to_solution(deploy_path: Path):
        """Gets the path to the solution behind the deploy_path. If folder is provided file called solution.py must
        live in the deploy_path.
        """
        if deploy_path.is_dir():
            path_to_solution = deploy_path.joinpath(
                DefaultValues.solution_default_name.value
            )
        else:
            path_to_solution = deploy_path

        return path_to_solution

    @staticmethod
    def retrieve_head_name(coordinates: ICoordinates):
        """Retrieves the branch (head) name for the merge request of the solution file."""
        return "_".join(
            [coordinates.group(), coordinates.name(), coordinates.version()]
        )

    @staticmethod
    def _create_merge_request(
        coordinates: ICoordinates,
        repo: Repo,
        file_paths,
        dry_run=False,
        push_option=None,
        email=None,
        username=None,
    ):
        """Creates a merge request to the catalog repository for the album object.

        Commits first the files given in the call, but will not include anything else than that.

        Args:
            file_paths:
                A list of files to include in the merge request. Can be relative to
                the catalog repository path or absolute.
            dry_run:
                Option to only show what would happen. No Merge request will be created.
            push_option:
                Push options for the repository.
            username:
                The git user to use. (Default: systems git configuration)
            email:
                The git email to use. (Default: systems git configuration)

        Raises:
            RuntimeError when no differences to the previous commit can be found.

        """
        if push_option is None:
            push_option = []

        # make a new branch and checkout
        new_head = create_new_head(repo, DeployManager.retrieve_head_name(coordinates))
        new_head.checkout()

        commit_msg = "Adding new/updated %s" % DeployManager.retrieve_head_name(
            coordinates
        )

        add_files_commit_and_push(
            new_head,
            file_paths,
            commit_msg,
            push=not dry_run,
            push_option_list=push_option,
            email=email,
            username=username,
            force=True,
        )

    @staticmethod
    def _push_directly(
        coordinates: ICoordinates,
        repo: Repo,
        files_to_remove: list,
        files_to_add: list,
        dry_run=False,
        push_option=None,
        email=None,
        username=None,
    ):
        if push_option is None:
            push_option = []

        # don't create a new branch. use reference branch from origin
        head = checkout_main(repo)

        commit_msg = "Adding new/updated %s" % DeployManager.retrieve_head_name(
            coordinates
        )
        remove_files(head, files_to_remove)

        add_files_commit_and_push(
            head,
            files_to_add,
            commit_msg,
            push=not dry_run,
            push_option_list=push_option,
            email=email,
            username=username,
            force=False,
        )

    @staticmethod
    def _create_yaml_file_in_local_src(active_solution: ISolution, solution_home: Path):
        """Creates a yaml file in the given repo for the given solution.

        Returns:
            The Path to the created markdown file.

        """
        yaml_path = solution_home.joinpath(
            DefaultValues.solution_yml_default_name.value
        )

        module_logger().debug("Writing yaml file to: %s..." % yaml_path)
        write_dict_to_yml(yaml_path, get_deploy_dict(active_solution))

        return yaml_path

    def _remove_db_entry_and_files(
        self, repo, coordinates, dry_run, push_options, git_email, git_name
    ):

        solution_root = str(
            self.album.configuration().get_solution_path_suffix_unversioned(coordinates)
        )
        db_index = DefaultValues.catalog_index_file_name.value

        init_msg = (
            "Would remove files %s, would update %s.."
            if dry_run
            else "Removing files %s, updating %s.."
        )
        module_logger().info(init_msg % (solution_root, db_index))

        if not dry_run:
            self._try_push_remove_tag(
                coordinates,
                repo,
                [solution_root],
                [db_index],
                push_options,
                git_email,
                git_name,
            )

    def _remove_db_entry_and_revert_files(
        self,
        repo,
        coordinates,
        previous_version_tag,
        dry_run,
        push_options,
        git_email,
        git_name,
    ):

        solution_root = str(
            self.album.configuration().get_solution_path_suffix_unversioned(coordinates)
        )
        db_index = DefaultValues.catalog_index_file_name.value

        init_msg = (
            "Would revert %s to %s, would update %s.."
            if dry_run
            else "Reverting %s to %s, updating %s.."
        )
        module_logger().info(init_msg % (solution_root, previous_version_tag, db_index))

        if not dry_run:
            revert(repo, previous_version_tag, [solution_root])
            self._try_push_remove_tag(
                coordinates, repo, [], [db_index], push_options, git_email, git_name
            )

    def _remove_db_entry_and_tag(
        self, repo, coordinates, dry_run, push_options, git_email, git_name
    ):
        db_index = DefaultValues.catalog_index_file_name.value

        init_msg = "Would update %s.." if dry_run else "Updating %s.."
        module_logger().info(init_msg % (db_index))

        if not dry_run:
            self._try_push_remove_tag(
                coordinates, repo, [], [db_index], push_options, git_email, git_name
            )

    def _try_push_remove_tag(
        self,
        coordinates,
        repo,
        files_to_remove,
        files_to_add,
        push_options,
        git_email,
        git_name,
    ):
        try:
            if push_options is None:
                push_options = []

            # don't create a new branch. use reference branch from origin
            head = checkout_main(repo)

            commit_msg = "Removing %s" % DeployManager.retrieve_head_name(coordinates)

            remove_tag(repo, as_tag(coordinates))

            remove_files(head, files_to_remove)

            add_files_commit_and_push(
                head,
                files_to_add,
                commit_msg,
                push=True,
                push_option_list=push_options,
                email=git_email,
                username=git_name,
                force=False,
            )
        except Exception as e:
            module_logger().error(
                "Pushing to catalog failed! Rolling back deployment..."
            )
            try:
                clean_repository(repo)
            finally:
                raise e



from album.core.api.model.environment import IEnvironment
from album.runner.core.api.model.coordinates import ICoordinates


class ScriptQueueEntry:
    def __init__(self, coordinates: ICoordinates, scripts, environment: IEnvironment):
        self.scripts = scripts
        self.environment = environment
        self.coordinates = coordinates



from abc import ABCMeta, abstractmethod

from album.core.api.controller.clone_manager import ICloneManager
from album.core.api.controller.collection.catalog_handler import ICatalogHandler
from album.core.api.controller.collection.collection_manager import ICollectionManager
from album.core.api.controller.collection.solution_handler import ISolutionHandler
from album.core.api.controller.deploy_manager import IDeployManager
from album.core.api.controller.environment_manager import IEnvironmentManager
from album.core.api.controller.event_manager import IEventManager
from album.core.api.controller.install_manager import IInstallManager
from album.core.api.controller.migration_manager import IMigrationManager
from album.core.api.controller.run_manager import IRunManager
from album.core.api.controller.script_manager import IScriptManager
from album.core.api.controller.search_manager import ISearchManager
from album.core.api.controller.state_manager import IStateManager
from album.core.api.controller.task_manager import ITaskManager
from album.core.api.controller.test_manager import ITestManager
from album.core.api.model.configuration import IConfiguration


class IAlbumController:
    __metaclass__ = ABCMeta

    def catalogs(self) -> ICatalogHandler:
        raise NotImplementedError

    @abstractmethod
    def solutions(self) -> ISolutionHandler:
        raise NotImplementedError

    @abstractmethod
    def configuration(self) -> IConfiguration:
        raise NotImplementedError

    @abstractmethod
    def environment_manager(self) -> IEnvironmentManager:
        raise NotImplementedError

    @abstractmethod
    def migration_manager(self) -> IMigrationManager:
        raise NotImplementedError

    @abstractmethod
    def script_manager(self) -> IScriptManager:
        raise NotImplementedError

    @abstractmethod
    def deploy_manager(self) -> IDeployManager:
        raise NotImplementedError

    @abstractmethod
    def install_manager(self) -> IInstallManager:
        raise NotImplementedError

    @abstractmethod
    def run_manager(self) -> IRunManager:
        raise NotImplementedError

    @abstractmethod
    def test_manager(self) -> ITestManager:
        raise NotImplementedError

    @abstractmethod
    def search_manager(self) -> ISearchManager:
        raise NotImplementedError

    @abstractmethod
    def clone_manager(self) -> ICloneManager:
        raise NotImplementedError

    @abstractmethod
    def state_manager(self) -> IStateManager:
        raise NotImplementedError

    @abstractmethod
    def collection_manager(self) -> ICollectionManager:
        raise NotImplementedError

    @abstractmethod
    def event_manager(self) -> IEventManager:
        raise NotImplementedError

    @abstractmethod
    def task_manager(self) -> ITaskManager:
        raise NotImplementedError

    @abstractmethod
    def close(self):
        raise NotImplementedError



from album.core.api.model.configuration import IConfiguration
from album.core.controller.conda_manager import CondaManager


class MambaManager(CondaManager):
    """Class for handling conda environments via mamba."""

    def __init__(self, configuration: IConfiguration):
        super().__init__(configuration)
        self._mamba_executable = self._configuration.mamba_executable()

    def _get_install_environment_executable(self):
        return self._mamba_executable



from abc import abstractmethod, ABCMeta
from typing import Optional, List

from album.core.api.model.database import IDatabase
from album.runner.core.api.model.coordinates import ICoordinates


class ICollectionIndex(IDatabase):
    __metaclass__ = ABCMeta

    class ICollectionSolution:
        __metaclass__ = ABCMeta

        @abstractmethod
        def setup(self) -> dict:
            raise NotImplementedError

        def internal(self) -> dict:
            raise NotImplementedError

    @abstractmethod
    def create(self):
        raise NotImplementedError

    @abstractmethod
    def update_name_version(self, name: str, version: str, close: bool = True):
        raise NotImplementedError

    @abstractmethod
    def get_name(self, close: bool = True) -> str:
        raise NotImplementedError

    @abstractmethod
    def get_version(self, close: bool = True) -> str:
        raise NotImplementedError

    @abstractmethod
    def is_empty(self, close: bool = True) -> bool:
        raise NotImplementedError

    @abstractmethod
    def insert_catalog(
        self, name, src, path, deletable, branch_name, catalog_type, close: bool = True
    ):
        raise NotImplementedError

    @abstractmethod
    def get_catalog(self, catalog_id, close: bool = True) -> dict:
        raise NotImplementedError

    @abstractmethod
    def get_catalog_by_name(self, catalog_name: str, close: bool = True) -> dict:
        raise NotImplementedError

    @abstractmethod
    def get_catalog_by_path(self, catalog_path: str, close: bool = True) -> dict:
        raise NotImplementedError

    @abstractmethod
    def get_catalog_by_src(self, catalog_src: str, close: bool = True) -> dict:
        raise NotImplementedError

    @abstractmethod
    def get_all_catalogs(self, close: bool = True) -> List[dict]:
        raise NotImplementedError

    @abstractmethod
    def remove_catalog(self, catalog_id, close: bool = True):
        raise NotImplementedError

    @abstractmethod
    def insert_solution(self, catalog_id, solution_attrs, close: bool = True):
        raise NotImplementedError

    @abstractmethod
    def get_all_solutions(self, close: bool = True) -> List[ICollectionSolution]:
        raise NotImplementedError

    @abstractmethod
    def get_all_installed_solutions_by_catalog(
        self, catalog_id, close: bool = True
    ) -> List[ICollectionSolution]:
        raise NotImplementedError

    @abstractmethod
    def get_solutions_by_catalog(
        self, catalog_id, close: bool = True
    ) -> List[ICollectionSolution]:
        raise NotImplementedError

    @abstractmethod
    def get_solution_by_hash(
        self, hash_value, close: bool = True
    ) -> Optional[ICollectionSolution]:
        raise NotImplementedError

    @abstractmethod
    def get_solution_by_collection_id(
        self, collection_id, close: bool = True
    ) -> Optional[ICollectionSolution]:
        raise NotImplementedError

    @abstractmethod
    def get_solution_by_doi(
        self, doi, close: bool = True
    ) -> Optional[ICollectionSolution]:
        raise NotImplementedError

    @abstractmethod
    def get_solution_by_catalog_grp_name_version(
        self, catalog_id, coordinates: ICoordinates, close: bool = True
    ) -> Optional[ICollectionSolution]:
        raise NotImplementedError

    @abstractmethod
    def get_solutions_by_grp_name_version(
        self, coordinates: ICoordinates, close: bool = True
    ) -> List[ICollectionSolution]:
        raise NotImplementedError

    @abstractmethod
    def get_solutions_by_name_version(
        self, name: str, version: str, close: bool = True
    ) -> List[ICollectionSolution]:
        raise NotImplementedError

    @abstractmethod
    def get_solutions_by_grp_name(
        self, group: str, name: str, close: bool = True
    ) -> List[ICollectionSolution]:
        raise NotImplementedError

    @abstractmethod
    def get_solutions_by_name(
        self, name: str, close: bool = True
    ) -> List[ICollectionSolution]:
        raise NotImplementedError

    @abstractmethod
    def get_recently_installed_solutions(
        self, close: bool = True
    ) -> List[ICollectionSolution]:
        raise NotImplementedError

    @abstractmethod
    def get_recently_launched_solutions(
        self, close: bool = True
    ) -> List[ICollectionSolution]:
        raise NotImplementedError

    @abstractmethod
    def get_unfinished_installation_solutions(self, close: bool = True):
        raise NotImplementedError

    @abstractmethod
    def update_solution(
        self,
        catalog_id,
        coordinates: ICoordinates,
        solution_attrs: dict,
        supported_attrs: list,
        close: bool = True,
    ):
        raise NotImplementedError

    @abstractmethod
    def add_or_replace_solution(
        self, catalog_id, coordinates: ICoordinates, solution_attrs, close: bool = True
    ):
        raise NotImplementedError

    @abstractmethod
    def remove_solution(
        self, catalog_id, coordinates: ICoordinates, close: bool = True
    ):
        raise NotImplementedError

    @abstractmethod
    def is_installed(self, catalog_id, coordinates: ICoordinates, close: bool = True):
        raise NotImplementedError

    @abstractmethod
    def insert_collection_collection(
        self,
        collection_id_parent,
        collection_id_child,
        catalog_id_parent,
        catalog_id_child,
        close=True,
    ):
        raise NotImplementedError

    @abstractmethod
    def remove_parent(self, collection_id, close=True):
        raise NotImplementedError



from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
from pathlib import Path
from typing import Optional, List, Generator

from git import Repo

from album.core.api.model.catalog_index import ICatalogIndex
from album.runner.core.api.model.coordinates import ICoordinates
from album.runner.core.api.model.solution import ISolution


class ICatalog:
    """Interface representing a Catalog.

    An album catalog contains solution files. These files are python scripts which can be interpreted by the album
    framework and implement a routine for solving a problem of any kind. The Catalog has an index where the solution
    files and its metadata are stored in a hierarchical way. This class brings all the functionality to resolve, add,
    remove solutions from a Catalog, whereas resolving refers to the act of looking up, if a solution exists in the
    catalog. A catalog can be local or remote. If the catalog is remote, solutions cannot be
    added or removed to/from it. (see deploy for context)

    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def catalog_id(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def src(self) -> str:
        """The source of the catalog. Gitlab/github link or path."""
        raise NotImplementedError

    @abstractmethod
    def version(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def index(self) -> Optional[ICatalogIndex]:
        raise NotImplementedError

    @abstractmethod
    def path(self) -> Path:
        """The path to the catalog cache."""
        raise NotImplementedError

    @abstractmethod
    def branch_name(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def is_deletable(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def is_cache(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def is_local(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def load_index(self):
        raise NotImplementedError

    @abstractmethod
    def solution_list_path(self) -> Path:
        """The path to the catalog solution list cache."""
        raise NotImplementedError

    @abstractmethod
    def index_file_path(self) -> Path:
        """The path to the catalog index cache."""
        raise NotImplementedError

    @abstractmethod
    def set_index_path(self, path) -> Path:
        """The path to the catalog index cache."""
        raise NotImplementedError

    @abstractmethod
    def type(self) -> str:
        """The type of the catalog."""
        raise NotImplementedError

    @abstractmethod
    def dispose(self):
        raise NotImplementedError

    @abstractmethod
    def update_index_cache_if_possible(self, tmp_dir):
        raise NotImplementedError

    @abstractmethod
    def update_index_cache(self, tmp_dir):
        raise NotImplementedError

    @abstractmethod
    def add(self, active_solution: ISolution, force_overwrite=False):
        raise NotImplementedError

    @abstractmethod
    def remove(self, coordinates: ICoordinates):
        raise NotImplementedError

    @contextmanager
    def retrieve_catalog(
        self, path=None, force_retrieve=False, update=True
    ) -> Generator[Repo, None, None]:
        yield
        raise NotImplementedError

    @abstractmethod
    def get_meta_information(self):
        raise NotImplementedError

    @abstractmethod
    def get_all_solution_versions(self, group: str, name: str) -> List[ISolution]:
        raise NotImplementedError

    @abstractmethod
    def set_catalog_id(self, catalog_id):
        raise NotImplementedError

    @abstractmethod
    def set_version(self, version):
        raise NotImplementedError

    @abstractmethod
    def get_version(self):
        raise NotImplementedError

    @abstractmethod
    def get_meta_file_path(self):
        raise NotImplementedError



import os
import platform
from pathlib import Path


class Link(type(Path())):
    _flavour = type(Path())._flavour
    _link = None

    def set_link(self, link):
        self._link = link
        return self

    def get_link(self):
        return self._link

    def dispose(self):
        if self._link:
            operation_system = platform.system().lower()
            link = os.path.normpath(self._link)
            if "windows" in operation_system:
                link += ".lnk"
            if os.path.exists(link) or os.path.islink(link):
                os.remove(link)



import os
from contextlib import contextmanager
from pathlib import Path
from typing import Optional, Tuple, List, Generator

import validators
from album.runner.core.api.model.coordinates import ICoordinates
from git import Repo, GitCommandError

from album.core.api.model.catalog import ICatalog
from album.core.api.model.catalog_index import ICatalogIndex
from album.core.model.catalog_index import CatalogIndex
from album.core.model.default_values import DefaultValues
from album.core.utils.operations.file_operations import copy, force_remove
from album.core.utils.operations.git_operations import (
    download_repository,
    clone_repository_sparse,
    checkout_files,
)
from album.core.utils.operations.resolve_operations import dict_to_coordinates
from album.core.utils.operations.solution_operations import get_deploy_dict
from album.runner import album_logging
from album.runner.core.api.model.solution import ISolution
from album.runner.core.model.solution import Solution

module_logger = album_logging.get_active_logger


def retrieve_index_files_from_src(
    src, tmp_dir: Path, branch_name="main"
) -> Tuple[Path, Path]:
    """Takes a src (path, or url) and retrieves the index files. Expects a git repository behind the src!"""
    tmp_dir = Path(tmp_dir)
    with clone_repository_sparse(src, branch_name, tmp_dir) as repo:
        # meta file - must be available
        checkout_files(repo, [DefaultValues.catalog_index_metafile_json.value])
        try:
            # db file - optional
            checkout_files(repo, [DefaultValues.catalog_index_file_name.value])
        except GitCommandError as e:
            # catalog index file does not have to be present for empty catalogs
            if "did not match any file" not in str(e):
                raise e
    index_src = tmp_dir.joinpath(DefaultValues.catalog_index_file_name.value)
    index_meta_src = tmp_dir.joinpath(DefaultValues.catalog_index_metafile_json.value)

    return index_src, index_meta_src


class Catalog(ICatalog):
    def __init__(
        self,
        catalog_id,
        name,
        path,
        src=None,
        deletable=True,
        branch_name="main",
        catalog_type="direct",
    ):
        """Init routine.

        Args:
            catalog_id:
                The id of the catalog in the collection.
            name:
                The name of the catalog.
            path:
                The absolute path to the catalog.
            src:
                The source of the catalog (Default: None)
                Can be a URL, a git-link, a path to a network-device or a path to any other storage system.
            deletable:
                Boolean to indicate whether the catalog is deletable or not. Relevant for a collection of catalogs.
            branch_name:
                When a git based catalog this attribute can be set to use other branches than the main branch (default)
            catalog_type:
                The type of the catalog. Either "direct" or "request". Important during deployment.

        """
        self._catalog_id = catalog_id
        self._name = name
        self._src = src
        self._version = None  # is set automatically with the index
        self._catalog_index: Optional[ICatalogIndex] = None
        self._path = Path(path)
        self._branch_name = branch_name

        self._is_deletable = deletable

        self._solution_list_path = self._path.joinpath(
            DefaultValues.catalog_solution_list_file_name.value
        )
        self._meta_file_path = self._path.joinpath(
            DefaultValues.catalog_index_metafile_json.value
        )
        self._index_file_path = self._path.joinpath(
            DefaultValues.catalog_index_file_name.value
        )
        self._type = catalog_type

        if self.is_local() and self._src:
            self._src = Path(self._src).absolute()

    def __eq__(self, other):
        return isinstance(other, ICatalog) and other.catalog_id() == self._catalog_id

    def __del__(self):
        self.dispose()

    def dispose(self):
        if self._catalog_index is not None:
            self._catalog_index.close()

    def is_cache(self):
        """Returns Boolean indicating whether the catalog is used for caching only."""
        return (
            self._src is None
            or self.is_local()
            and self._path.exists()
            and os.path.samefile(str(self._src), self._path)
        )

    def is_local(self):
        """Returns Boolean indicating whether the catalog is remote or local."""
        return not self._src or (
            not validators.url(str(self._src)) and Path(self._src).exists()
        )

    def update_index_cache_if_possible(self, tmp_dir):
        try:
            self.update_index_cache(tmp_dir)
        except GitCommandError as e:
            module_logger().warning("Could not refresh index. Git command failed:")
            module_logger().warning(e)
            return False
        except Exception as e:
            module_logger().warning("Could not refresh index. Unknown reason!")
            module_logger().warning(e)
            return False

        return True

    def update_index_cache(self, tmp_dir):
        if self.is_cache():
            return False

        index_available = self._update_index_cache(tmp_dir)

        if not index_available:
            self.dispose()

        return True

    def add(self, active_solution: ISolution, force_overwrite=False):
        solution_attrs = get_deploy_dict(active_solution)

        if active_solution.setup().doi:
            solution_attrs["doi"] = active_solution.setup().doi
            if (
                self._catalog_index.get_solution_by_doi(solution_attrs["doi"])
                is not None
            ):
                if force_overwrite:
                    module_logger().warning("Solution already exists! Overwriting...")
                else:
                    raise RuntimeError("DOI already exists in catalog! Aborting...")

        if active_solution.setup().deposit_id:
            solution_attrs["deposit_id"] = active_solution.setup().deposit_id

        lookup_solution = self._catalog_index.get_solution_by_coordinates(
            active_solution.coordinates()
        )
        if lookup_solution:
            if force_overwrite:
                module_logger().warning("Solution already exists! Overwriting...")
            else:
                raise RuntimeError("Solution already exists in catalog! Aborting...")

        self._catalog_index.update(active_solution.coordinates(), solution_attrs)
        self._catalog_index.save()
        self._catalog_index.export(self._solution_list_path)

    def remove(self, coordinates: ICoordinates):
        solution_entry = self._catalog_index.remove_solution_by_group_name_version(
            coordinates
        )
        if solution_entry:
            self._catalog_index.export(self._solution_list_path)
        else:
            module_logger().warning("Solution not found! Doing nothing...")

    def _update_index_cache(self, tmp_dir):
        repo_dir = Path(tmp_dir).joinpath("repo")
        try:
            src, meta_src = retrieve_index_files_from_src(
                self.src(), repo_dir, branch_name=self.branch_name()
            )
            return self._copy_index_to_cache(src, meta_src)
        finally:
            force_remove(repo_dir)

    def _copy_index_to_cache(self, db_file, meta_file):
        # check if meta information valid, only then continue
        if not meta_file.exists():
            raise FileNotFoundError("Could not find file %s..." % meta_file)

        module_logger().debug(
            "Copying meta information from %s to %s..."
            % (meta_file, self._meta_file_path)
        )
        copy(meta_file, self._meta_file_path)

        if db_file.exists():
            module_logger().debug(
                "Copying index from %s to %s..." % (db_file, self._index_file_path)
            )
            copy(db_file, self._index_file_path)
        else:
            if self._index_file_path.exists():
                force_remove(self._index_file_path)
                return False
            else:
                module_logger().debug("Index file of the catalog does not exist yet...")
                return False
        return True

    @contextmanager
    def retrieve_catalog(
        self, path=None, force_retrieve=False, update=True
    ) -> Generator[Repo, None, None]:
        if self.is_cache():
            raise RuntimeError("Cannot retrieve a cache catalog as no source exists!")

        path = Path(path) if path else self._path

        module_logger().debug(
            "Trying to retrieve catalog %s to the path %s..." % (self._name, str(path))
        )
        repo = download_repository(
            self._src, str(path), force_download=force_retrieve, update=update
        )

        yield repo
        repo.close()

    def get_meta_information(self):
        return {"name": self._name, "version": self._version, "type": self._type}

    def get_all_solution_versions(self, group: str, name: str) -> List[Solution]:
        versions = self._catalog_index.get_all_solution_versions(group, name)
        res = []
        for version in versions:
            res.append(Solution(attrs=version))
        return res

    def load_index(self):
        self._catalog_index = CatalogIndex(self._name, self._index_file_path)

    def catalog_id(self) -> int:
        return self._catalog_id

    def name(self) -> str:
        return self._name

    def src(self) -> Path:
        return self._src

    def version(self) -> str:
        return self._version

    def index(self) -> ICatalogIndex:
        return self._catalog_index

    def path(self) -> Path:
        return self._path

    def branch_name(self) -> str:
        return self._branch_name

    def is_deletable(self) -> bool:
        return self._is_deletable

    def solution_list_path(self) -> Path:
        return self._solution_list_path

    def index_file_path(self) -> Path:
        return self._index_file_path

    def set_index_path(self, path):
        self._index_file_path = Path(path)

    def type(self) -> str:
        return self._type

    def set_catalog_id(self, catalog_id):
        self._catalog_id = catalog_id

    def set_version(self, version):
        self._version = version

    def get_version(self):
        return self._version

    def get_meta_file_path(self):
        return self._meta_file_path



import queue
import threading
from threading import Thread

from album.core.api.controller.task_manager import ITaskManager
from album.core.api.model.task import ITask
from album.core.model.task import LogHandler, Task
from album.runner import album_logging

module_logger = album_logging.get_active_logger


class TaskManager(ITaskManager):
    server_queue = None

    num_fetch_threads = 3
    tasks = {}

    task_count = 0
    workers_initialized = False

    def __init__(self):
        self.tasks = {}
        self.task_count = 0
        self.server_queue = queue.Queue()
        self.workers_initialized = False

    def _initialize_workers(self):
        self.workers_initialized = True
        current_thread = threading.current_thread().ident
        module_logger().info(
            f"TaskManager: Initializing {self.num_fetch_threads} worker threads..."
        )
        for i in range(self.num_fetch_threads):
            worker = Thread(target=self._run_queue_entry, args=(i, current_thread))
            worker.setDaemon(True)
            worker.start()

    def get_task(self, task_id):
        return self.tasks[task_id]

    def get_status(self, task: ITask):
        return {
            "status": task.status().name,
            "records": self._records_to_json(task.log_handler().records()),
        }

    @staticmethod
    def _records_to_json(records):
        res = []
        for record in records:
            res.append(
                {
                    "asctime": None
                    if not hasattr(record, "asctime")
                    else str(record.asctime),
                    "name": None if not hasattr(record, "name") else str(record.name),
                    "levelname": None
                    if not hasattr(record, "levelname")
                    else str(record.levelname),
                    "msg": None if not hasattr(record, "msg") else str(record.msg),
                }
            )
        return res

    def _finish_queue(self):
        self.server_queue.join()
        # TODO add timeout parameter and do something like the following if timeout is set:
        # timeout = 10  # waiting for 10 seconds for queue to finish
        # stop = time() + timeout
        # while self.server.task_manager.server_queue.unfinished_tasks and time() < stop:
        #     sleep(1)

    def create_and_register_task(self, method, args) -> str:
        task = Task(method, args)
        return self.register_task(task)

    def register_task(self, task: ITask) -> str:
        if not self.workers_initialized:
            self._initialize_workers()
        task.set_id(str(self.task_count))
        module_logger().info(f"TaskManager: registering task {task.id()}")
        task.set_status(ITask.Status.WAITING)
        self.task_count += 1
        self.tasks[task.id()] = task
        self.server_queue.put(task)
        return task.id()

    def _run_queue_entry(self, i, parent_thread):
        album_logging.configure_logging(
            "worker" + str(i), parent_thread_id=parent_thread
        )
        while True:
            task = self.server_queue.get()
            self._handle_task(task)
            self.server_queue.task_done()

    def _handle_task(self, task: ITask):
        module_logger().info(f"TaskManager: starting task {task.id()}...")
        logger = album_logging.configure_logging("task" + str(task.id()))
        handler = LogHandler()
        task.set_log_handler(handler)
        logger.addHandler(handler)
        task.set_status(ITask.Status.RUNNING)
        try:
            self._run_task(task)
            task.set_status(ITask.Status.FINISHED)
        except Exception as e:
            logger.error(e)
            task.set_status(ITask.Status.FAILED)
        logger.removeHandler(handler)
        album_logging.pop_active_logger()
        module_logger().info(f"TaskManager: finished task {task.id()}.")

    def finish_tasks(self):
        self.server_queue.join()

    @staticmethod
    def _run_task(task: ITask):
        return task.method()(*task.args())



from abc import ABCMeta, abstractmethod

from album.core.api.model.catalog import ICatalog
from album.core.api.model.collection_index import ICollectionIndex


class IMigrationManager:
    """Interface handling the migration from Indices of different versions. Responsible for execution of SQL migration!

    Tries to always migrate the catalog index or collection index to the current framework version
    able to deal with the database.

    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def migrate_collection_index(
        self, collection_index: ICollectionIndex, initial_version
    ):
        """Migrates a collection index database to the target version."""
        raise NotImplementedError

    @abstractmethod
    def migrate_catalog_collection_db(
        self, collection_index_path, curr_version, target_version
    ):
        """Migrates a given collection index to the target version."""
        raise NotImplementedError

    @abstractmethod
    def migrate_catalog_index_db(
        self, catalog_index_path, curr_version, target_version
    ):
        """Migrates the index of a given catalog to the current framework version if possible."""
        raise NotImplementedError

    @abstractmethod
    def load_index(self, catalog: ICatalog):
        """Loads the index from file or src. If a file and src exists routine tries to update the index."""
        raise NotImplementedError

    @abstractmethod
    def refresh_index(self, catalog: ICatalog) -> bool:
        """Routine to refresh the catalog index. Downloads or copies the index_file."""
        raise NotImplementedError

    @abstractmethod
    def migrate_solution_attrs(self, attrs):
        raise NotImplementedError



import logging
import sys

from album.core.utils.operations.view_operations import (
    get_logger_name_minimizer_filter,
    get_logging_formatter,
    get_message_filter,
)
from album.runner.album_logging import (
    set_loglevel,
    debug_settings,
    LogLevel,
    push_active_logger,
)


def configure_root_logger(
    log_format: str = None, log_format_time: str = None, log_level: LogLevel = None
):
    if not log_level:
        log_level = LogLevel(debug_settings())
    logger = logging.getLogger("album")
    logger.setLevel(log_level.name)
    # create console handler and set level to debug
    # ToDo: different handlers necessary? e.g. logging additional into a file?
    ch = logging.StreamHandler(stream=sys.stdout)
    ch.setLevel(log_level.name)

    # add formatter to ch
    ch.setFormatter(get_logging_formatter(fmt=log_format, time=log_format_time))
    ch.addFilter(get_logger_name_minimizer_filter())
    ch.addFilter(get_message_filter())

    # add ch to logger
    logger.addHandler(ch)
    set_loglevel(log_level)
    push_active_logger(logger)


def add_logging_level(levelName, levelNum, methodName=None):
    """
    Copied from https://stackoverflow.com/a/35804945

    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example
    -------
    >>> add_logging_level('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5

    """
    if not methodName:
        methodName = levelName.lower()

    if hasattr(logging, levelName):
        raise AttributeError("{} already defined in logging module".format(levelName))
    if hasattr(logging, methodName):
        raise AttributeError("{} already defined in logging module".format(methodName))
    if hasattr(logging.getLoggerClass(), methodName):
        raise AttributeError("{} already defined in logger class".format(methodName))

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def logForLevel(self, message, *args, **kwargs):
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)

    def logToRoot(message, *args, **kwargs):
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)



from abc import ABCMeta, abstractmethod

from album.core.api.model.event import IEvent
from album.runner.core.api.model.coordinates import ICoordinates


class IEventManager:
    """Interface handling events."""

    __metaclass__ = ABCMeta

    @abstractmethod
    def add_listener(self, event_name, callback, coordinates: ICoordinates = None):
        """Add callback method for specific event"""
        raise NotImplementedError

    @abstractmethod
    def remove_listener(self, event_name, callback, coordinates: ICoordinates = None):
        """Add callback method for specific event"""
        raise NotImplementedError

    def publish(self, event: IEvent, coordinates: ICoordinates = None):
        """Publish an event to anyone listening to events with the same name"""
        raise NotImplementedError



import pathlib
from typing import Optional, Union

from album.core.api.controller.controller import IAlbumController
from album.core.api.model.catalog import ICatalog
from album.core.api.model.collection_index import ICollectionIndex
from album.core.api.model.collection_solution import ICollectionSolution
from album.core.api.model.configuration import IConfiguration
from album.core.api.model.event import IEvent
from album.core.api.model.task import ITask
from album.core.controller.album_controller import AlbumController
from album.core.utils.core_logging import configure_root_logger
from album.runner.album_logging import pop_active_logger, LogLevel
from album.runner.core.api.model.script_creator import IScriptCreator
from album.runner.core.api.model.solution import ISolution


class Album:
    class Builder:
        _base_cache_path: Optional[Union[str, pathlib.Path]] = None
        _log_format: str = None
        _log_format_time: str = None
        _log_level: LogLevel = None

        def base_cache_path(
            self, base_cache_path: Union[str, pathlib.Path]
        ) -> "Album.Builder":
            self._base_cache_path = base_cache_path
            return self

        def log_format(self, log_format: str) -> "Album.Builder":
            self._log_format = log_format
            return self

        def log_format_time(self, log_format_time: str) -> "Album.Builder":
            self._log_format_time = log_format_time
            return self

        def log_level(self, log_level: LogLevel) -> "Album.Builder":
            self._log_level = log_level
            return self

        def build(self) -> "Album":
            _controller = AlbumController(self._base_cache_path)
            configure_root_logger(
                log_format=self._log_format,
                log_format_time=self._log_format_time,
                log_level=self._log_level,
            )
            return Album(_controller, logger_pushed=True)

    def __init__(self, album_controller: IAlbumController, logger_pushed=False) -> None:
        self._controller = album_controller
        self.logger_pushed = logger_pushed

    def __del__(self):
        self.close()

    def resolve(self, resolve_solution: str) -> ICollectionSolution:
        return self._controller.collection_manager().resolve_and_load(resolve_solution)

    def resolve_installed(self, resolve_solution: str) -> ICollectionSolution:
        return self._controller.collection_manager().resolve_installed_and_load(
            resolve_solution
        )

    def load_or_create_collection(self):
        self._controller.collection_manager().load_or_create()

    def get_index_as_dict(self):
        return self._controller.collection_manager().get_index_as_dict()

    def get_catalogs_as_dict(self):
        return self._controller.collection_manager().catalogs().get_all_as_dict()

    def get_catalog_by_name(self, name) -> ICatalog:
        return self._controller.catalogs().get_by_name(name)

    def get_catalog_by_src(self, src) -> ICatalog:
        return self._controller.catalogs().get_by_src(src)

    def get_collection_index(self) -> ICollectionIndex:
        return self._controller.collection_manager().get_collection_index()

    def test(self, solution_to_resolve: str, args=None):
        return self._controller.test_manager().test(solution_to_resolve, args)

    def load(self, path) -> Optional[ISolution]:
        return self._controller.state_manager().load(path)

    def search(self, keywords):
        """Searches through album catalogs to find closest matching solution."""
        return self._controller.search_manager().search(keywords)

    def run(self, solution_to_resolve: str, argv=None, run_async=False):
        return self._run_async(
            self._controller.run_manager().run,
            (solution_to_resolve, False, argv),
            run_async,
        )

    def install(self, solution_to_resolve: str, argv=None, run_async=False):
        return self._run_async(
            self._controller.install_manager().install,
            (solution_to_resolve, argv),
            run_async,
        )

    def uninstall(
        self, solution_to_resolve: str, rm_dep=False, argv=None, run_async=False
    ):
        """Removes a solution from the disk. Thereby uninstalling its environment and deleting all its downloads.

        Args:
            argv:
                Arguments which should be appended to the script call
            solution_to_resolve:
                The solution to remove
            rm_dep:
                Boolean to indicate whether to remove parents too.

        """
        return self._run_async(
            self._controller.install_manager().uninstall,
            (solution_to_resolve, rm_dep, argv),
            run_async,
        )

    def deploy(
        self,
        deploy_path: str,
        catalog_name: str,
        dry_run: bool,
        push_options=None,
        git_email: str = None,
        git_name: str = None,
        force_deploy: bool = False,
        changelog: str = "",
    ):
        """Function corresponding to the `deploy` subcommand of `album`.

        Generates the yml for a album and creates a merge request to the catalog only
        including the yaml and solution file.

        Args:
            force_deploy:
                Force overwrites a existing solution during deployment. Only for local catalogs.
            deploy_path:
                Path to a directory or a file.
                If directory: Must contain "solution.py" file.
            catalog_name:
                The catalog to deploy to. Either specify via argument in deploy-call, via url in solution or use
                default catalog.
            dry_run:
                When set, prepares deployment in local src of the catlog (creating zip, docker, yml),
                but not adding to the catalog src.
            push_options:
                Push options for the catalog repository.
            git_email:
                The git email to use. (Default: systems git configuration)
            git_name:
                The git user to use. (Default: systems git configuration)
            changelog:
                The change associated with this version of a solution compared to the last version.

        """
        return self._controller.deploy_manager().deploy(
            deploy_path=deploy_path,
            catalog_name=catalog_name,
            dry_run=dry_run,
            push_options=push_options,
            git_email=git_email,
            git_name=git_name,
            force_deploy=force_deploy,
            changelog=changelog,
        )

    def undeploy(
        self,
        solution_to_resolve: str,
        catalog_name: str,
        dry_run: bool,
        push_options=None,
        git_email: str = None,
        git_name: str = None,
    ):
        """Function corresponding to the `undeploy` subcommand of `album`.

        Removes the solution from the given catalog.

        Args:
            solution_to_resolve:
                Solution identifier which should be removed (group:name:version).
            catalog_name:
                The catalog to remove the solution from.
            dry_run:
                When set, prepares undeploy in local src of the catalog,
                but not actually removing it the catalog src.
            push_options:
                Push options for the catalog repository.
            git_email:
                The git email to use. (Default: systems git configuration)
            git_name:
                The git user to use. (Default: systems git configuration)

        """
        return self._controller.deploy_manager().undeploy(
            solution_to_resolve=solution_to_resolve,
            catalog_name=catalog_name,
            dry_run=dry_run,
            push_options=push_options,
            git_email=git_email,
            git_name=git_name,
        )

    def clone(self, path: str, target_dir: str, name: str) -> None:
        """
        Function corresponding to the `clone` subcommand of `album`.

        Args:
            path: the source of the clone command - a solution (group:name:version, path, or URL to file) or a catalog
                template string (i.e. template:catalog)
            target_dir: the directory where the cloned solution or catalog will be added to
            name: the name of the solution or catalog to be created

        """
        return self._controller.clone_manager().clone(path, target_dir, name)

    def close(self):
        if self.logger_pushed:
            pop_active_logger()
        self._controller.close()

    def run_solution_script(
        self, resolve_result: ICollectionSolution, script: IScriptCreator
    ):
        self._controller.script_manager().run_solution_script(resolve_result, script)

    def upgrade(self, catalog_name=None, dry_run=False, override=False):
        return (
            self._controller.collection_manager()
            .catalogs()
            .update_collection(catalog_name, dry_run, override)
        )

    def update(self, catalog_name=None):
        return self._controller.collection_manager().catalogs().update_any(catalog_name)

    def add_catalog(self, catalog_src):
        return self._controller.collection_manager().catalogs().add_by_src(catalog_src)

    def remove_catalog_by_src(self, catalog_src):
        return (
            self._controller.collection_manager()
            .catalogs()
            .remove_from_collection_by_src(catalog_src)
        )

    def remove_catalog_by_name(self, catalog_src):
        return (
            self._controller.collection_manager()
            .catalogs()
            .remove_from_collection_by_name(catalog_src)
        )

    def configuration(self) -> IConfiguration:
        return self._controller.configuration()

    def is_installed(self, solution_to_resolve: str):
        resolve_result = self.resolve(solution_to_resolve)
        return self._controller.solutions().is_installed(
            resolve_result.catalog(), resolve_result.coordinates()
        )

    def load_catalog_index(self, catalog: ICatalog):
        return self._controller.migration_manager().load_index(catalog)

    def add_event_listener(self, event_name, callback, solution_id=None):
        return self._controller.event_manager().add_listener(
            event_name, callback, solution_id
        )

    def remove_event_listener(self, event_name, callback, solution_id=None):
        return self._controller.event_manager().remove_listener(
            event_name, callback, solution_id
        )

    def publish_event(self, event: IEvent):
        return self._controller.event_manager().publish(event)

    def get_task_status(self, task_id) -> ITask.Status:
        """Get the status of a task managed by the task manager."""
        task = self._controller.task_manager().get_task(task_id)
        if task is None:
            raise LookupError("Task with id %s not found." % task_id)
        return self._controller.task_manager().get_status(task)

    def create_and_register_task(self, method, args) -> str:
        return self._controller.task_manager().create_and_register_task(method, args)

    def finish_tasks(self):
        return self._controller.task_manager().finish_tasks()

    def _run_async(self, method, args, run_async=False):
        if run_async:
            return self._controller.task_manager().create_and_register_task(
                method, args
            )
        else:
            return method(*args)
